"""
Add ability for Rainbow to keep track of their own history.
"""
from ...imports import *

__all__ = [
    "_setup_history",
    "_record_history_entry",
    "_remove_last_history_entry",
    "_create_history_entry",
    "history",
]


def _setup_history(self):
    """
    Record the first history entry in the log.

    Parameters
    ----------
    h : dict
        A dictionary containing keys `name` as the name
        of the action and `inputs` as its keyword inputs.
    """
    self.metadata["history"] = []


def _record_history_entry(self, h):
    """
    Record a history entry in the log.

    Parameters
    ----------
    h : dict
        A dictionary containing keys `name` as the name
        of the action and `inputs` as its keyword inputs.
    """
    self.metadata["history"].append(h)


def _remove_last_history_entry(self):
    """
    Remove the most recent history entry from the log
    """
    try:
        self.metadata["history"].pop()
    except KeyError:
        self.metadata["history"] = []


def represent_as_copypasteable(x):
    """
    Represent a particular quantity as a copy-pastable string
    within the chromatic import environment.

    Parameters
    ----------
    x : any
        The thing to represent

    Returns
    -------
    r : str
        The thing represented as a string
    """
    if isinstance(x, u.Quantity):
        value = represent_as_copypasteable(x.value)
        unit = x.unit
        return f"u.Quantity({value})*u.Unit('{unit}')"
    elif isinstance(x, np.ndarray):
        return f"np.{repr(x)}"
    elif isinstance(x, slice):
        return (
            f"{x.start}:{x.stop}:{x.step}".replace("None", "")
            .replace("::", ":")
            .replace("::", ":")
        )
    else:
        return repr(x)


def _create_history_entry(self, name, inputs={}):
    """
    Create a history entry.

    Parameters
    ----------
    name : str
        The name of the action being applied.
    inputs : dict
        Dictionary of all inputs.

    Returns
    -------
    h : dict
        A dictionary containing keys `name` as the name
        of the action and `inputs` as its keyword inputs.
    """

    # remove "self" from the list of inputs
    inputs.pop("self", None)

    # remove None inputs
    to_remove = [k for k in inputs if inputs[k] is None]
    for k in to_remove:
        inputs.pop(k, None)

    # create a dictionary to store function call components
    if name == "__getitem__":
        w = represent_as_copypasteable(inputs["i_wavelength"])
        t = represent_as_copypasteable(inputs["i_time"])
        call = f"[{w},{t}]"
    elif name in "+-*/":
        try:
            other = f'\n{textwrap.indent(inputs["other"].history(), " ")}'
        except AttributeError:
            other = represent_as_copypasteable(inputs["other"])
        call = f"{name}{other}"
    else:
        list_of_arguments = [
            f"{k}={represent_as_copypasteable(v)}" for k, v in inputs.items()
        ]
        arguments_as_string = "\n   " + ",\n   ".join(list_of_arguments)

        if "Rainbow" in name:
            call = f"{name}({arguments_as_string})"
        else:
            call = f".{name}({arguments_as_string})"

    return call


def history(self):
    """
    Return a summary of the history of actions that have gone into this object.


    Returns
    -------
    history : ?
        The history of actions, in the requested format

    """

    calls = self.metadata["history"]
    return "(\n" + "\n".join(calls) + "\n)"



from ....imports import *

__all__ = ["plot_noise_comparison_in_bins"]


def plot_noise_comparison_in_bins(
    self,
    ax=None,
    cmap=None,
    vmin=None,
    vmax=None,
    expected=True,
    measured_errorbarkw={},
    expected_plotkw={},
    filename=None,
    ylim=[1e-5, 1e-1],
    **kw,
):
    """
    ax : matplotlib.axes.Axes
        The axes into which the plot should be drawn.
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The wavelength at the bottom of the cmap.
    vmax : astropy.units.Quantity
        The wavelength at the top of the cmap.
    errorbar_kw : dict

    **kw : dictionary
        Additional keywords are passed to `.get_measured_scatter_in_bins`.
        Please see the docstring for that method for options.

    """

    # figure out where the plot should be drawn
    if ax is None:
        ax = plt.subplot()
    plt.sca(ax)

    # make sure the color map is set up
    self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

    # calculate the binned down scatters
    x = self.get_measured_scatter_in_bins(**kw)

    # loop through wavelengths
    for i, w in enumerate(self.wavelength):
        color = self.get_wavelength_color(w)
        ekw = dict(marker="o", color=color)
        ekw.update(**measured_errorbarkw)
        plt.errorbar(x["N"], x["scatters"][i], x["uncertainty"][i], **ekw)
        if expected:
            pkw = dict(color=color, alpha=0.2)
            pkw.update(**expected_plotkw)
            plt.plot(x["N"], x["expectation"][i], **pkw)

    # set good plot appearance default
    plt.yscale("log")
    plt.xscale("log")
    plt.ylim(*ylim)
    plt.xlim(1, np.nanmax(x["N"]))
    plt.xlabel("# of Times in Bin")

    t_unit = u.minute
    twin_ax = ax.twiny()

    plt.sca(twin_ax)
    dt = x["dt"].to(t_unit).value
    plt.plot(dt, x["expectation"][i], alpha=0)
    plt.xlim(np.nanmin(dt), np.nanmax(dt))
    plt.xscale("log")
    plt.xlabel(f"Duration of Bin ({t_unit.to_string()})")

    plt.sca(ax)
    # plt.title(self.get("title"))

    if filename is not None:
        self.savefig(filename)
    return ax



from ....imports import *
from ..utilities import *

__all__ = ["plot_one_wavelength_with_models", "animate_with_models"]


def _create_title(s):
    """
    A helper function to make better titles.
    """
    if s == "model":
        return "Relative Flux"
    elif "model" in s:
        return "Flux (" + s.replace("_", " ").title().replace(" Model", "") + ")"
    else:
        return s.title()


def plot_one_wavelength_with_models(
    self,
    i_wavelength,
    models=["model", "systematics_model", "planet_model"],
    vlimits_data=[None, None],
    vspan_residuals=None,
    panelsize=(3.25, 2),
    orientation="vertical",
    errorbar=True,
    text=True,
    data_errorbarkw={},
    model_plotkw={},
    wavelength_textkw={},
    cmap=None,
    vmin=None,
    vmax=None,
    w_unit="micron",
    t_unit="day",
    label="inside",
    label_textkw={"color": "black"},
    animation=False,
    minimum_acceptable_ok=1e-10,
    **kw,
):
    """
    Produce a multi-panel plot figure comparing
    the data to model components. The first panel
    will be the data ('flux'), the last panel will
    be the residuals ('flux' - 'model'), and the middle
    panel(s) will be the model(s) connecting them.

    Parameters
    ----------
    i_wavelength : int
        The index of the wavelength to plot.
    models : list
        The list of model quantities to display,
        each in its own panel. A popular option might be:
        `models = ['systematics_model', 'planet_model']
        Whatever you include here does not affect the
        values of the Residuals panel, which are always
        exactly 'flux' - 'model'.
    vlimits_data : list
        The value limits for the data and model colormap,
        specified as `vlimits_data = [vmin, vmax]`.
        If `None` or `[None, None]`, the limits will be
        replaced with the [1,99] percentiles of the flux.
    vspan_residuals : float
        The distance away from zero where the colormap
        for the residuals will cut off. If `None`, the
        span will be set to 3 standard deviations.
    panelsize : tuple
        The size of each panel of the multipanel plot.
        It should probably be wider than it is tall.
    orientation : string
        The direction in which the panels should spread out.
        Options include ['vertical', 'horizontal'].
    errorbar : boolean
        Should we plot errorbars?
    text : boolean
        Should we label each wavelength?
    data_errorbarkw : dict
        A dictionary of keywords passed to `plt.errorbar`
        so you can have more detailed control over the data
        appearance. Common keyword arguments might include:
        `[alpha, elinewidth, color, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.errorbar.html
    model_plotkw : dict
        A dictionary of keywords passed to `plt.plot`
        so you can have more detailed control over the model
        appearance. Common keyword arguments might include:
        `[alpha, clip_on, zorder, marker, markersize,
          linewidth, linestyle, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html
    wavelength_textkw : dict
        A dictionary passed to `plt.text` for the wavelength label
        so you can have more detailed control over the text
        appearance. Common keyword arguments might include:
        `[alpha, backgroundcolor, color, fontfamily, fontsize,
          fontstyle, fontweight, rotation, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.text.html
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    t_unit : str, astropy.unit.Unit
        The unit for plotting times.
    label : bool
        Should we add (a), (b), (c) labels to the panels,
        and where? False or None give no labels, 'inside'
        sets them in the upper left inside corner, 'outside'
        sets them above the upper left corner.
    label_textkw : dict
        A dictionary passed to `plt.text` for the (a), (b), (c) label
        so you can have more detailed control over the text
        appearance. Common keyword arguments might include:
        `[alpha, backgroundcolor, color, fontfamily, fontsize,
          fontstyle, fontweight, rotation, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.text.html
    animation : bool
        Are we in the middle of an animation
        (so shouldn't make a new figure)?
    **kw : dict
        All additional keywords will be ignored.
    """

    # make sure the units work
    w_unit, t_unit = u.Unit(w_unit), u.Unit(t_unit)

    # make sure we use only models that exist
    models_that_exist = []
    for m in models:
        if self.get(m) is None:
            warnings.warn(f"'{m}' doesn't exist and will be skipped.")
        else:
            models_that_exist.append(m)
    models = models_that_exist

    # make sure vertical limits are set for data and model
    percentiles = [0, 100]
    if vlimits_data == None or vlimits_data == [None, None]:
        vlimits_data = np.percentile(self.flux, percentiles)
    elif vlimits_data[0] == None:
        vlimits_data[0] = np.percentile(self.flux, percentiles[0])
    elif vlimits_data[1] == None:
        vlimits_data[1] = np.percentile(self.flux, percentiles[1])

    # make sure vertical limits are set for the residuals
    if vspan_residuals is None:
        vspan_residuals = np.nanstd(self.residuals) * 5

    # make sure that the wavelength-based colormap is defined
    self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

    # set up the panels
    N = len(models) + 1
    if orientation == "vertical":
        rows, columns = N, 1
        sharex, sharey = "all", "none"
        vspan_residual_ratio = np.maximum(
            vspan_residuals / (vlimits_data[1] - vlimits_data[0]), 0.5
        )
        height_ratios = [1] * (N - 1) + [vspan_residual_ratio]
        gridspec_kw = dict(height_ratios=height_ratios)
        figsize = (panelsize[0], np.sum(height_ratios) * panelsize[1])
    elif orientation == "horizontal":
        rows, columns = 1, N
        sharex, sharey = "all", "all"
        gridspec_kw = dict()
        figsize = (panelsize[0] * N, panelsize[1])
    else:
        raise ValueError(
            f"Please choose an orientation from ['vertical', 'horizontal']."
        )

    # allow animation by making an option to not remake figure and axes
    setup_keys = ["fi", "ax", "vlimits_data", "vspan_residuals"]
    try:
        # try to pull the pre-existing figures and axes (if making animation)
        fi = self._animate_with_models_setup["fi"]
        ax = self._animate_with_models_setup["ax"]
        vlimits_data = self._animate_with_models_setup["vlimits_data"]
        vspan_residuals = self._animate_with_models_setup["vspan_residuals"]
        assert animation == True
        for a in ax:
            a.cla()
    except (AttributeError, AssertionError):
        # otherwise, create  new figure and axes
        fi, ax = plt.subplots(
            rows,
            columns,
            figsize=figsize,
            dpi=300,
            sharex=sharex,
            sharey=sharey,
            gridspec_kw=gridspec_kw,
            constrained_layout=True,
            facecolor="white",
        )
        self._animate_with_models_setup = {}
        for k in setup_keys:
            self._animate_with_models_setup[k] = vars()[k]

    # get the automatic color for this particular wavelength
    w = self.wavelength[i_wavelength]
    color = self.get_wavelength_color(w)

    # set the default xlabel
    xlabel = f"Time ({t_unit.to_string('latex_inline')})"

    # set defaults for the model line plots
    plotkw = dict(marker=None, linewidth=1, zorder=1, color=color, alpha=0.5)
    plotkw.update(**model_plotkw)

    # set default for the data errorbar plots
    default_markersize = plt.matplotlib.rcParams["lines.markersize"]
    errorbarkw = dict(
        color=color,
        linewidth=0,
        elinewidth=int(errorbar),
        marker="o",
        # markersize=int(errorbar == False) * default_markersize,
        markeredgecolor="none",
        alpha=0.5,
        zorder=-1,
    )
    errorbarkw.update(**data_errorbarkw)

    # pull out the raw data and complete model
    data_x, data_y, data_sigma = self.get_ok_data_for_wavelength(
        i_wavelength,
        x="time",
        y="flux",
        sigma="uncertainty",
        minimum_acceptable_ok=minimum_acceptable_ok,
    )
    model_x, model_y, _ = self.get_ok_data_for_wavelength(
        i_wavelength,
        x="time",
        y="model",
        sigma="uncertainty",
        minimum_acceptable_ok=minimum_acceptable_ok,
    )

    # loop through models to include
    for i, (k, a) in enumerate(zip(models + ["residuals"], ax)):

        # choose the model we want to isolate
        if k == "residuals":
            this_model_x, this_model_y, _ = self.get_ok_data_for_wavelength(
                i_wavelength,
                x="time",
                y="ones",
                minimum_acceptable_ok=minimum_acceptable_ok,
            )
        else:
            this_model_x, this_model_y, _ = self.get_ok_data_for_wavelength(
                i_wavelength, x="time", y=k, minimum_acceptable_ok=minimum_acceptable_ok
            )

        # point to the appropriate plotting axes
        plt.sca(a)

        # plot the data
        plt.errorbar(
            data_x.to(t_unit),
            data_y / model_y * this_model_y,
            yerr=data_sigma,
            **errorbarkw,
        )

        # plot the model
        plt.plot(this_model_x.to(t_unit), this_model_y, **plotkw)

        # set the axis labels and limits
        if orientation == "vertical":
            a.set_ylabel(_create_title(k))
            if k == "residuals":
                a.set_xlabel(xlabel)
                a.set_ylim(1 - vspan_residuals, 1 + vspan_residuals)
            else:
                a.set_ylim(*vlimits_data)
        elif orientation == "horizontal":
            a.set_title(_create_title(k))
            if i == 0:
                a.set_ylabel("Relative Flux")
            a.set_ylim(*vlimits_data)
            a.set_xlabel(xlabel)

    # add text labels for the wavelength
    textkw = dict(
        x=0.98,
        y=0.04,
        va="bottom",
        ha="right",
        transform=ax[0].transAxes,
        color=color,
    )
    textkw.update(**wavelength_textkw)
    textkw.update(s=f"{w.to_value(w_unit):.2f} {w_unit.to_string('latex_inline')}")
    if text:
        ax[0].text(**textkw)

    # add (a), (b), (c) labels to the panels, if desired
    if label:
        _add_panel_labels(ax[:], preset=label, **label_textkw)

    # set xlimits
    ax[0].set_xlim(self.time.to_value(t_unit).min(), self.time.to_value(t_unit).max())
    plt.sca(ax[0])


def animate_with_models(
    self,
    filename="animated-lightcurves-with-models.gif",
    fps=5,
    bitrate=None,
    dpi=None,
    orientation="horizontal",
    **kw,
):

    self.plot_one_wavelength_with_models(
        0, animation=False, orientation=orientation, **kw
    )

    filename = self._label_plot_file(filename)

    # initialize the animator
    writer, displayer = _get_animation_writer_and_displayer(
        filename=filename, fps=fps, bitrate=bitrate
    )

    # set up to save frames directly into the animation
    figure = self._animate_with_models_setup["fi"]
    with writer.saving(figure, filename, dpi or figure.get_dpi()):

        # loop over exposures
        for i in tqdm(range(self.nwave), leave=False):

            self.plot_one_wavelength_with_models(
                i, animation=True, orientation=orientation, **kw
            )
            # save this snapshot to a movie frame
            writer.grab_frame()

    # close the figure that was created
    plt.close(figure)

    # display the animation
    from IPython.display import display

    try:
        display(displayer(filename, embed=True))
    except TypeError:
        display(displayer(filename))



from ....imports import *

__all__ = ["plot_histogram"]


def plot_histogram(
    self,
    i_wavelength,
    ax=None,
    quantity="flux",
    color="auto",
    cmap=None,
    vmin=None,
    vmax=None,
    orientation="vertical",
    expected=False,
    expected_plotkw={},
    **kw,
):
    """
    Plots a histogram of the flux value for one wavelength of a rainbow.

    Parameters
    ----------
    i_wavelength : integer
        The wavelength row we want to plot.
    ax : matplotlib.axes.Axes
        The axes into which the plot should be drawn.
    quantity : string
        The quantity for which we want a histogram.
        Currently available options are ['flux', 'residuals']
    color : string
        The color for the histogram. If 'auto', guess from wavelength.
    expected : Boolean
        Allows user to choose whether an expected normal distribution for the
        data will be plotted over the histogram.
    expected_plotkw : dict
        Keywords to pass to `plt.plot` for plotting the expected distribution.
    **kw : dictionary
        All additional keywords will be passed to `plt.hist`.
        Please see the documentation for that function for options.
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.hist.html)

    Returns
    -------
        A plotted histogram of flux values for one wavelength, with flux bin
        values on the x-axis and the number of occurrences (= histogram) of
        flux values within a particular bin on the y-axis.

    """

    if ax is None:
        ax = plt.subplot()
    plt.sca(ax)

    # importing norm so we can plot normal distribution over histogram
    from scipy.stats import norm

    # pick the color of the histogram based on the wavelength
    if color == "auto":
        self._make_sure_cmap_is_defined()
        color = self.get_wavelength_color(self.wavelength[i_wavelength])

    # get the quantity for which we want to plot the histogram
    assert quantity in ["flux", "residuals"]
    time, flux, uncertainty = self.get_ok_data_for_wavelength(i_wavelength, y=quantity)

    # plotting histogram of row 'i' of data (wavelength 'i')
    histkw = dict(alpha=0.5)
    histkw.update(**kw)
    plt.hist(flux, color=color, density=True, orientation=orientation, **histkw)
    if orientation == "vertical":
        plt.xlabel(f"{quantity}")
        plt.ylabel(f"P({quantity})")
    elif orientation == "horizontal":
        plt.ylabel(f"{quantity}")
        plt.xlabel(f"P({quantity})")

    # option to plot expected normal distribution over histogram
    if expected == True:
        # mu (middle of the normal distribution) corresponds to the middle
        # of the flux array
        if quantity == "residuals":
            mu = 0
        else:
            mu = np.median(flux)

        # expected std corresponds to the middle of our true uncertainty
        std = np.median(uncertainty)

        # setting min/max values for scale of normal distribution
        nsigma = 4
        xmin = np.minimum(np.min(flux), mu - nsigma * std)
        xmax = np.maximum(np.max(flux), mu + nsigma * std)

        x = np.linspace(xmin, xmax, 500)
        p = norm.pdf(x, mu, std)
        plotkw = dict(color=color, alpha=0.5, linewidth=3)
        plotkw.update(**expected_plotkw)
        if orientation == "vertical":
            plt.plot(x, p, **plotkw)
        elif orientation == "horizontal":
            plt.plot(p, x, **plotkw)



"""
Define a reader for Néstor's NIRISS extracted spectra .npy files.

This was kludged together to read the Stage 2 outputs from
https://stsci.app.box.com/s/tyg3qqd85601gkbw5koowrx0obekeg0m/folder/154382588636
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_espinoza"]


def from_espinoza(rainbow, filepath):
    """
    Populate a Rainbow from a file in the espinoza format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load. It should be a glob-friendly
        string like `.../*_order1.npy` that points to both a
        `spectra_order1.npy` and a `wavelengths_order1.npy` file.
    """

    # read in two files, one for the flux, one for the wavelength
    spectra_filename, wavelengths_filename = sorted(glob.glob(filepath))
    assert ("spectra") in spectra_filename
    assert ("wavelength") in wavelengths_filename
    spectra = np.load(spectra_filename)
    wavelengths = np.load(wavelengths_filename)

    # populate a 1D array of wavelengths (with astropy units of length)
    rainbow.wavelike["wavelength"] = wavelengths * u.micron * 1

    # populate a 1D array of times (with astropy units of time)
    times = np.arange(spectra.shape[0]) * u.minute
    warnings.warn("The times are totally made up!")
    rainbow.timelike["time"] = times * 1

    # populate a 2D (row = wavelength, col = array of fluxes
    rainbow.fluxlike["flux"] = spectra[:, 0, :].T * 1
    rainbow.fluxlike["uncertainty"] = spectra[:, 1, :].T * 1



from ...imports import *

__all__ = [
    "setup_wavelength_colors",
    "_make_sure_cmap_is_defined",
    "get_wavelength_color",
]


def setup_wavelength_colors(self, cmap=None, vmin=None, vmax=None, log=None):
    """
    Set up a color map and normalization function for
    colors datapoints by their wavelengths.

    Parameters
    ----------
    cmap : str, matplotlib.colors.Colormap
        The color map to use.
    vmin : astropy.units.Quantity
        The wavelength at the bottom of the cmap.
    vmax : astropy.units.Quantity
        The wavelength at the top of the cmap.
    log : bool
        If True, colors will scale with log(wavelength).
        If False, colors will scale with wavelength.
        If None, the scale will be guessed from the internal wscale.
    """

    # populate the cmap object
    self.cmap = cm.get_cmap(cmap)

    vmin = vmin
    if vmin is None:
        vmin = np.nanmin(self.wavelength)
    vmax = vmax
    if vmax is None:
        vmax = np.nanmax(self.wavelength)

    if (self.wscale in ["log"]) or (log == True):
        self.norm = col.LogNorm(
            vmin=vmin.to("micron").value, vmax=vmax.to("micron").value
        )
    if (self.wscale in ["?", "linear"]) or (log == False):
        self.norm = col.Normalize(
            vmin=vmin.to("micron").value, vmax=vmax.to("micron").value
        )


def _make_sure_cmap_is_defined(self, cmap=None, vmin=None, vmax=None):
    """
    A helper function that can be called at the start of
    any plot that that's using wavelength-colors to make
    sure that the wavelength-based colormap has been
    defined.

    Parameters
    ----------
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap.
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap.
    """

    if hasattr(self, "cmap"):
        if (cmap is not None) or (vmin is not None) or (vmax is not None):
            if (
                (cmap != self.get("cmap"))
                and (vmin != self.get("vmin"))
                and (vmax != self.get("vmax"))
            ):
                warnings.warn(
                    """
                It looks like you're trying to set up a new custom
                cmap and/or wavelength normalization scheme. You
                should be aware that a cmap has already been defined
                for this object; if you're visualizing the same
                rainbow in different ways, we strongly suggest
                that you not change the cmap or normalization
                between them, for visual consistency.
                """
                )
        else:
            return
    self.setup_wavelength_colors(cmap=cmap, vmin=vmin, vmax=vmax)


def get_wavelength_color(self, wavelength):
    """
    Determine the color corresponding to one or more wavelengths.

    Parameters
    ----------
    wavelength : astropy.units.Quantity
        The wavelength value(s), either an individual
        wavelength or an array of N wavelengths.

    Returns
    -------
    colors : np.array
        An array of RGBA colors [or an (N,4) array].
    """
    w_unitless = wavelength.to("micron").value
    normalized_w = self.norm(w_unitless)
    return self.cmap(normalized_w)



from ....imports import *

__all__ = ["get_average_spectrum"]


def get_average_spectrum(self):
    """
    Return a average_spectrum of the star, averaged over all times.

    Parameters
    ----------

    Returns
    -------
    average_spectrum : np.array (wavelike)
    """
    return self.get_average_spectrum_as_rainbow().flux[:, 0]



"""
Define a reader for chromatic .rainbow.FITS files.
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_rainbow_FITS"]


def from_rainbow_FITS(rainbow, filepath):
    """
    Populate a Rainbow from a file in the rainbow_FITS format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the file to load.
    """

    # open the FITS file
    hdu_list = fits.open(filepath)

    # load the header into metadata
    h = hdu_list["primary"].header
    for k in h:
        if k.lower() in [
            "name",
            "wscale",
            "tscale",
            "signal_to_noise",
            "history",
        ]:  # KLUDGE!
            rainbow.metadata[k.lower()] = h[k]
        elif k not in ["SIMPLE", "BITPIX", "NAXIS", "EXTEND"]:
            rainbow.metadata[k] = h[k]

    # load into the three core dictionaries
    for e in ["fluxlike", "wavelike", "timelike"]:
        table = Table.read(hdu_list[e])
        for k in table.colnames:
            vars(rainbow)[e][k] = table[k].quantity * 1



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_kirk_fitted_light_curves", "from_kirk_stellar_spectra"]


def from_kirk_fitted_light_curves(self, filepath):
    """
    Populate a Rainbow from a file in James Kirk's fitted
    light curve format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load. This is interpreted to
        be pointing to a directory called `wb_lcs/` that's
        filled with wavelength-binned light curves. It will
        expect there to also be a transmission spectrum
        file called 'transmission_spectrum.data' at the
        same directory level as 'wb_lcs'.

            some-directory/
                wb_lcs/
                transmission_spectrum.dat
    """

    # the directory that contains the light curve
    lc_directory = os.path.abspath(filepath)

    # the expected location of the transmission spectrum
    transmission_spectrum_filepath = lc_directory.replace(
        "wb_lcs", "transmission_spectrum.dat"
    )

    # complain if the spectrum isn't found
    try:
        transmission_spectrum_file = glob.glob(transmission_spectrum_filepath)[0]
    except IndexError:
        raise ValueError(
            f"""
        The `from_kirk_fitted_lightcurves` reader is expecting to find
        {transmission_spectrum_filepath}
        but it's not there 🤨.
        """
        )

    # load the transmission spectrum
    transmission_spectrum = ascii.read(
        transmission_spectrum_file,
        names=[
            "wavelength",
            "wavelength_width",
            "radius_ratio",
            "radius_ratio_uncertainty_upper",
            "radius_ratio_uncertainty_lower",
        ],
    )

    # populate a 1D array of wavelengths (with astropy units of length)
    for k in transmission_spectrum.colnames:
        if "wavelength" in k:
            unit = u.micron
        else:
            unit = 1
        self.wavelike[k] = transmission_spectrum[k].data * unit
    self.wavelike["wavelength_lower"] = (
        self.wavelike["wavelength"] - self.wavelike["wavelength_width"] / 2
    )
    self.wavelike["wavelength_upper"] = (
        self.wavelike["wavelength"] + self.wavelike["wavelength_width"] / 2
    )

    # the filenames for the individual light curve files
    lc_filenames = sorted(glob.glob(os.path.join(lc_directory, "model_tab_wb*.dat")))
    assert len(lc_filenames) == self.nwave

    # loop through the files and load them into arrays (faster than tables)
    lcs = []
    for f in tqdm(lc_filenames, leave=False):
        lc = np.loadtxt(f)
        lcs.append(lc)
    names = [
        "time",
        "flux",
        "uncertainty",
        "rescaled_uncertainty",
        "model",
        "planet_model",
        "systematics_model",
        "residuals",
    ]

    # extract the time and the fluxlike quantities
    for i, k in enumerate(names):
        if k == "time":
            self.timelike["time"] = lcs[0][:, 0] * u.day
            for lc in lcs:
                assert np.shape(lc)[0] == self.ntime
        else:
            self.fluxlike[k] = np.array([lc[:, i] for lc in lcs])


def from_kirk_stellar_spectra(self, filepath):
    """
    Populate a Rainbow from a file in James Kirk's extracted
    stellar spectra format. This expects a file path pointing
    to a wb_lcs`

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load. This is interpreted to
        be pointing to a file called '*flux_resampled_*.pickle',
        and expects there to be at the same directory level two
        other files called '*error_resampled_*.pickle' and
        'wavelengths.pickle'.

            some-directory/
                star1_flux_resampled_unsmoothed.pickle
                star1_error_resampled_unsmoothed.pickle
                wvl_solution.pickle
                BJD_TDB_time.pickle
    """

    # load the flux pickle
    flux_file = filepath
    self.fluxlike["flux"] = pickle.load(open(flux_file, "rb")).T

    # load the uncertainty pickle
    uncertainty_file = filepath.replace("_flux_resampled", "_error_resampled")
    assert uncertainty_file != flux_file
    self.fluxlike["uncertainty"] = pickle.load(open(uncertainty_file, "rb")).T

    # load the wavelength pickle
    wavelength_file = flux_file.replace(
        os.path.basename(flux_file), "wvl_solution.pickle"
    )
    self.wavelike["wavelength"] = pickle.load(open(wavelength_file, "rb")) * u.micron

    # load the wavelength pickle
    time_file = flux_file.replace(os.path.basename(flux_file), "BJD_TDB_time.pickle")
    bjd_mjd = pickle.load(open(time_file, "rb"))
    astropy_times = Time(bjd_mjd, format="mjd", scale="tdb")
    self.set_times_from_astropy(astropy_times, is_barycentric=True)



from ....imports import *

__all__ = ["get_measured_scatter_in_bins"]


def get_measured_scatter_in_bins(
    self, ntimes=2, nbins=4, method="standard-deviation", minimum_acceptable_ok=1e-10
):
    """
    Get measured scatter in time bins of increasing sizes.
    For uncorrelated Gaussian noise, the scatter should
    decrease as 1/sqrt(N), where N is the number points
    in a bin.

    Parameters
    ----------
    ntimes : int
        How many times should be binned together? Binning will
        continue recursively until fewer that nbins would be left.
    nbins : int
        What's the smallest number of bins that should be used to
        calculate a scatter? The absolute minimum is 2.
    method : string
        What method to use to obtain measured scatter. Current options are 'MAD', 'standard-deviation'.
    minimum_acceptable_ok : float
        The smallest value of `ok` that will still be included.
        (1 for perfect data, 1e-10 for everything but terrible data, 0 for all data)

    Returns
    -------
    scatter : np.array (wavelike)
        The scatter wavelike array.
    """

    from ...rainbow import Rainbow

    if "remove_trends" in self.history():
        warnings.warn(
            f"""
        The `remove_trends` function was applied to this `Rainbow`,
        making it very plausible that some long-timescale signals
        and/or noise have been suppressed. Be suspicious of binned
        scatters on long timescales.
        """
        )

    # create a simplified rainbow so we don't waste time binning
    simple = Rainbow(
        time=self.time,
        wavelength=self.wavelength,
        flux=self.flux,
        uncertainty=self.uncertainty,
        ok=self.ok,
    )

    # loop through binning until done
    binnings = [simple]
    N = [1]
    while binnings[-1].ntime > ntimes * nbins:
        binnings.append(
            binnings[-1].bin(ntimes=ntimes, minimum_acceptable_ok=minimum_acceptable_ok)
        )
        N.append(N[-1] * ntimes)

    scatters = [b.get_measured_scatter(method=method) for b in binnings]
    expectation = [b.get_expected_uncertainty() for b in binnings]
    uncertainty_on_scatters = (
        scatters
        / np.sqrt(2 * (np.array([b.ntime for b in binnings]) - 1))[:, np.newaxis]
    )
    dt = [np.median(np.diff(b.time)) for b in binnings]

    return dict(
        N=np.array(N),
        dt=u.Quantity(dt),
        scatters=np.transpose(scatters),
        expectation=np.transpose(expectation),
        uncertainty=np.transpose(uncertainty_on_scatters),
    )
    # (see equation 3.48 of Sivia and Skilling for the uncertainty on sigma)



from ..rainbows import *
from .setup_tests import *


def test_shift():
    N = 1000
    w = np.linspace(0.6, 0.7, N) * u.micron
    f = u.Quantity(np.zeros(N))
    for i in range(3):
        w0 = np.random.uniform(0.6, 0.7) * u.micron
        sigma = np.random.uniform(0.001, 0.01) * u.micron
        f += np.exp(-0.5 * ((w - w0) / sigma) ** 2)

    for v in [0, -1e4, 1e4] * u.km / u.s:
        print(f"Testing for v={v}")
        unshifted = SimulatedRainbow(wavelength=w, star_flux=f).inject_noise()
        shifted = unshifted.shift(v)
        shifted_and_then_shifted_back = shifted.shift(-v)

        fi, ax = plt.subplots(
            1, 3, sharex=True, sharey=True, figsize=(8, 4), constrained_layout=True
        )
        unshifted.imshow(ax=ax[0])
        shifted.imshow(ax=ax[1])
        shifted_and_then_shifted_back.imshow(ax=ax[2])

        assert np.all(
            np.isclose(unshifted.wavelength, shifted_and_then_shifted_back.wavelength)
        )

        if v.to_value("km/s") > 0:
            print(f"The velocity {v} should cause a redshift.")
            assert np.all(shifted.wavelength > unshifted.wavelength)
        elif v.to_value("km/s") == 0:
            print(f"The velocity {v} should cause no shift.")
            print("   It does!")
        elif v.to_value("km/s") < 0:
            print(f"The velocity {v} should cause a blueshift.")
            assert np.all(shifted.wavelength < unshifted.wavelength)
        print()



from ...imports import *

__all__ = ["shift"]


def shift(self, velocity=0 * u.km / u.s):
    """
    Doppler shift the wavelengths of this Rainbow.
    Positive velocities make wavelengths longer (redshift).
    Negative velocities make wavelengths shorter (bluesfhit).

    Parameters
    ----------
    velocity : astropy.units.Quantity
        the systemic velocity in question,
        with units of velocity (for example, u.km/u.s)
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("shift", locals())

    # create a new copy of this rainbow
    new = self._create_copy()

    # get the speed of light from astropy constants
    lightspeed = con.c.to("km/s")  # speed of light in km/s

    # calculate beta and make sure the units cancel
    beta = (velocity / lightspeed).decompose()

    # apply wavelength shift
    new_wavelength = new.wavelength * np.sqrt((1 + beta) / (1 - beta))
    new.wavelike["wavelength"] = new_wavelength

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new object
    return new



from ...imports import *

__all__ = ["imshow"]


def imshow(
    self,
    ax=None,
    quantity="flux",
    xaxis="time",
    w_unit="micron",
    t_unit="day",
    colorbar=True,
    aspect="auto",
    mask_ok=True,
    color_ok="tomato",
    alpha_ok=0.8,
    vmin=None,
    vmax=None,
    filename=None,
    **kw,
):
    """
    Paint a 2D image of flux as a function of time and wavelength,
    using `plt.imshow` where pixels will have constant size.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    quantity : str
        The fluxlike quantity to imshow.
        (Must be a key of `rainbow.fluxlike`).
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    t_unit : str, astropy.unit.Unit
        The unit for plotting times.
    colorbar : bool
        Should we include a colorbar?
    aspect : str
        What aspect ratio should be used for the imshow?
    mask_ok : bool
        Should we mark which data are not OK?
    color_ok : str
        The color to be used for masking data points that are not OK.
    alpha_ok : float
        The transparency to be used for masking data points that are not OK.
    kw : dict
        All other keywords will be passed on to `plt.imshow`,
        so you can have more detailed control over the plot
        appearance. Common keyword arguments might include:
        `[cmap, norm, interpolation, alpha, vmin, vmax]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html
    """

    # self.speak(f'imshowing')
    if ax is None:
        ax = plt.subplot()

    # get units
    w_unit, t_unit = u.Unit(w_unit), u.Unit(t_unit)

    # make sure some wavelength and time edges are defined
    self._make_sure_wavelength_edges_are_defined()
    self._make_sure_time_edges_are_defined()

    # set up the wavelength extent
    try:
        wmin = self.wavelength_lower[0].to_value(w_unit)
        wmax = self.wavelength_upper[-1].to_value(w_unit)
    except AttributeError:
        wmin, wmax = None, None

    if (self.wscale == "linear") and (wmin is not None) and (wmax is not None):
        wlower, wupper = wmin, wmax
        wlabel = f"{self._wave_label} ({w_unit.to_string('latex_inline')})"
    elif self.wscale == "log" and (wmin is not None) and (wmax is not None):
        wlower, wupper = np.log10(wmin), np.log10(wmax)
        wlabel = (
            r"log$_{10}$" + f"[{self._wave_label}/({w_unit.to_string('latex_inline')})]"
        )
    else:
        message = f"""
        The wavelength scale for this rainbow is '{self.wscale}',
        and there are {self.nwave} wavelength centers and
        {len(self.wavelike.get('wavelength_lower', []))} wavelength edges defined.

        It's hard to imshow something with a wavelength axis
        that isn't linearly or logarithmically uniform, or doesn't
        at least have its wavelength edges defined. We're giving up
        and just using the wavelength index as the wavelength axis.

        If you want a real wavelength axis, one solution would
        be to bin your wavelengths to a more uniform grid with
        `rainbow.bin(R=...)` (for logarithmic wavelengths) or
        `rainbow.bin(dw=...)` (for linear wavelengths)
        """
        warnings.warn(message)
        wlower, wupper = -0.5, self.nwave - 0.5
        wlabel = "Wavelength Index"

    # set up the time extent
    try:
        tmin = self.time_lower[0].to_value(t_unit)
        tmax = self.time_upper[-1].to_value(t_unit)
    except AttributeError:
        tmin, tmax = None, None
    if (self.tscale == "linear") and (tmin is not None) and (tmax is not None):
        tlower, tupper = tmin, tmax
        tlabel = f"{self._time_label} ({t_unit.to_string('latex_inline')})"
    elif self.tscale == "log" and (tmin is not None) and (tmax is not None):
        tlower, tupper = np.log10(tmin), np.log10(tmax)
        tlabel = (
            r"log$_{10}$" + f"[{self._time_label}/({t_unit.to_string('latex_inline')})]"
        )
    else:
        message = f"""
        The time scale for this rainbow is '{self.tscale}',
        and there are {self.ntime} time centers and
        {len(self.timelike.get('time_lower', []))} time edges defined.

        It's hard to imshow something with a time axis
        that isn't linearly or logarithmically uniform, or doesn't
        at least have its time edges defined. We're giving up
        and just using the time index as the time axis.

        If you want a real time axis, one solution would
        be to bin your times to a more uniform grid with
        `rainbow.bin(dt=...)` (for linear times).
        """
        warnings.warn(message)
        tlower, tupper = -0.5, self.ntime - 0.5
        tlabel = "Time Index"

    def get_2D(k):
        """
        A small helper to get a 2D quantity. This is a bit of
        a kludge to help with weird cases of duplicate keys
        (for example where 'wavelength' might appear in both
        `wavelike` and `fluxlike`).
        """
        z = self.get(k)
        if np.shape(z) == self.shape:
            return z
        else:
            return self.fluxlike.get(k, None)

    if xaxis.lower()[0] == "t":
        self.metadata["_imshow_extent"] = [tlower, tupper, wupper, wlower]
        xlabel, ylabel = tlabel, wlabel
        z = get_2D(quantity)
        ok = get_2D("ok")
    elif xaxis.lower()[0] == "w":
        self.metadata["_imshow_extent"] = [wlower, wupper, tupper, tlower]
        xlabel, ylabel = wlabel, tlabel
        z = get_2D(quantity).T
        ok = get_2D("ok").T
    else:
        warnings.warn(
            "Please specify either `xaxis='time'` or `xaxis='wavelength'` for `.plot()`"
        )

    # figure out a good shared color limits (unless already supplied)
    vmin = vmin or np.nanpercentile(u.Quantity(z.flatten()).value * 1.0, 1)
    vmax = vmax or np.nanpercentile(u.Quantity(z.flatten()).value * 1.0, 99)

    # define some default keywords
    imshow_kw = dict(interpolation="nearest", vmin=vmin, vmax=vmax)
    imshow_kw.update(**kw)
    with quantity_support():
        plt.sca(ax)

        # create an overlaying mask of which data are OK or not
        if mask_ok:
            okimshow_kw = dict(**imshow_kw)
            okimshow_kw.update(
                cmap=one2another(
                    bottom=color_ok,
                    top=color_ok,
                    alpha_bottom=alpha_ok,
                    alpha_top=0,
                ),
                zorder=10,
                vmin=0,
                vmax=1,
            )
            plt.imshow(
                ok,
                extent=self.metadata["_imshow_extent"],
                aspect=aspect,
                origin="upper",
                **okimshow_kw,
            )
        plt.imshow(
            z,
            extent=self.metadata["_imshow_extent"],
            aspect=aspect,
            origin="upper",
            **imshow_kw,
        )
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        if colorbar:
            plt.colorbar(
                ax=ax,
                label=u.Quantity(z).unit.to_string("latex_inline"),
            )
        plt.title(self.get("title"))

    if filename is not None:
        self.savefig(filename)
    return ax



"""
Read multiple exposures of a single order of LCO/NRES spectra.
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_nres"]


def from_nres(rainbow, filepath, order=52):
    """
    Populate a Rainbow from a file in the NRES format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the file to load.

    order : float
        The order to extract.
        Acceptable values span 52 to 119
    """

    # read in your file, however you like
    filenames = glob.glob(filepath)
    filenames = np.sort(filenames)

    order_index = order - 52  # the 0th order is order 52

    for i, f in tqdm(enumerate(filenames), leave=False):

        # open this
        hdu = fits.open(f)

        # get the time associated with this observation
        date = hdu["PRIMARY"].header["MJD-OBS"]
        date_bjd = date + 2400000.5

        # get the science spectrum
        science_fiber_id = hdu["PRIMARY"].header["SCIFIBER"]
        is_science = hdu["SPECTRUM"].data["FIBER"] == science_fiber_id
        data = hdu["SPECTRUM"].data[is_science]
        this_order = data[order_index]
        order = this_order["ORDER"]
        ok = this_order["MASK"] == 0

        # these should be the good data for the particular order in question
        wavelengths = np.array(this_order["WAVELENGTH"]) * u.angstrom  # [mask]
        normalized_fluxes = this_order["NORMFLUX"]  # [mask]
        normalized_uncertainties = np.abs(this_order["NORMUNCERTAINTY"])  # [mask])

        sort = np.argsort(wavelengths.value)
        wavelengths = wavelengths[sort]
        fluxes = normalized_fluxes[sort]
        errors = normalized_uncertainties[sort]

        if i == 0:
            ntimes = len(filenames)
            nwaves = len(wavelengths)
            for k in ["wavelength_2d", "flux", "uncertainty", "ok"]:
                rainbow.fluxlike[k] = np.zeros((nwaves, ntimes))
            rainbow.fluxlike["wavelength_2d"] *= u.micron
            rainbow.fluxlike["ok"] = rainbow.fluxlike["ok"].astype(np.bool)
            rainbow.timelike["time"] = np.zeros(ntimes) * u.day

        # populate a 1D array of times (with astropy units of time)
        rainbow.timelike["time"][i] = date_bjd * u.day * 1

        # populate a 2D (row = wavelength, col = time, value = wavelength)
        rainbow.fluxlike["wavelength_2d"][:, i] = wavelengths * 1

        # populate a 2D (row = wavelength, col = time) array of fluxes
        rainbow.fluxlike["flux"][:, i] = fluxes * 1

        # populate a 2D (row = wavelength, col = time) array of uncertainties
        rainbow.fluxlike["uncertainty"][:, i] = errors * 1

        # populate a 2D (row = wavelength, col = time) array of ok
        rainbow.fluxlike["ok"][:, i] = ok * 1

    # populate a 1D array of wavelengths (with astropy units of length)
    rainbow.wavelike["wavelength"] = np.nanmedian(
        rainbow.fluxlike["wavelength_2d"], axis=rainbow.timeaxis
    )

    # add some warnings if there's any funny business
    if len(filenames) == 0:
        warnings.warn(
            f"""
        There are no files of that name in this folder!
        """
        )



from ..utilities import *
from ....imports import *

__all__ = ["plot_average_spectrum"]


def plot_average_spectrum(self, filename=None, **kw):
    """
    Plot the weighted average spectrum as a function of wavelength.

    Parameters
    ----------
    **kw : dict
        Additional keywords will be passed onward to the helper
        function `._scatter_timelike_or_wavelike`. Please see its
        docstrings for options about plot appearance and layout.
    """
    y = self.get_average_spectrum()
    self._scatter_timelike_or_wavelike(
        x=self.wavelength,
        y=y,
        ylabel=f"Flux ({_get_unit_string(y)})",
        **kw,
    )
    if filename is not None:
        self.savefig(filename)



from ....imports import *


def get_times_as_astropy(self, time=None, format=None, scale=None, is_barycentric=None):
    """
    Convert times from a Rainbow into an astropy Time object.

    Parameters
    ----------
    time : u.Quantity
        The time-like Quantity to be converted.
        If None (default), convert the time values in `self.time`
        If another time-like Quantity, convert those values.
    format : str
        The time format to supply to astropy.time.Time.
        If None (default), format will be pulled from
        `self.metadata['time_details']['format']`
    scale : str
        The time scale to supply to astropy.time.Time.
        If None (default), scale will be pulled from
        `self.metadata['time_details']['scale']`
    is_barycentric : bool
        Are the times already measured relative to the
        Solar System barycenter? This is mostly for warning
        the user that it's not.
        If None (default), is_barycentric will be pulled from
        `self.metadata['time_details']['is_barycentric']`

    Returns
    -------
    astropy_time : astropy.time.Time
        The times as an astropy Time object.
    """

    # take times from self or from the keyword
    if time is None:
        time = self.time

    # give a format warning
    format = format or self.get("time_format")
    if format is None:
        warnings.warn(
            f"""
        `.metadata['time_details']['format']` is not set,
        nor was a `format=` keyword argument provided.

        Since `.time` is already an astropy Quantity,
        this is likely a question of whether the format is
        'jd' or 'mjd' (= 'jd' - 2400000.5) or something else.
        If you get this wrong, you might be lost in time!

        For more about astropy.Time formats, please see:
        https://docs.astropy.org/en/stable/time/index.html#time-format
        """
        )

    # give a scale warning
    scale = scale or self.get("time_scale")
    if scale is None:
        now = Time.now()
        differences_string = ""
        for s in now.SCALES:
            dt = ((getattr(now, s).jd - now.tdb.jd) * u.day).to(u.second)
            differences_string += f"{s:>15} - tdb = {dt:10.6f}\n"
        warnings.warn(
            f"""
        .metadata['time_details']['scale'] is not set,
        nor was a `scale=` keyword argument provided.

        The main question is whether the time scale is 'tdb'
        (Barycentric Dynamical Time) or something close to it,
        or 'utc' or something close to it. The differences
        between these options, at {now.utc.iso} (UTC), are:
        \n{differences_string}
        If you get this wrong, you might be lost in time!

        For more about astropy.Time scales, please see:
        https://docs.astropy.org/en/stable/time/index.html#time-scale
        """
        )

    # give some barycenter warnings
    is_barycentric = is_barycentric or self.get("time_is_barycentric")
    if is_barycentric == True and "ut" in scale.lower():
        warnings.warn(
            f"""
        barycentic={is_barycentric} and scale={scale}
        It's a deeply weird combination to have a barycentric
        time measured at the Solar System barycentric but in
        Earth's leap-second-based UTC system. Please consider
        checking your time details.
        """
        )
    if is_barycentric != True:
        warnings.warn(
            f"""
        The returned time is not known to be measured relative
        to the Solar System barycenter. It's probably therefore
        measured from Earth or the position of your telescope,
        but please be warned that the timing of very distant event
        (like exoplanet transits) might be off by up to about
        8 minutes (= the light travel time between Earth + Sun).
        """
        )

    # generate astropy Time array
    astropy_time = Time(self.time, format=format, scale=scale)

    # do a check that the values aren't really weird
    if (astropy_time.min().decimalyear < 1000) or (
        astropy_time.max().decimalyear > 3000
    ):
        warnings.warn(
            f"""
        The times, which span
        jd={astropy_time.min().jd} to jd={astropy_time.max().jd}
        don't seem likely to be within the range of modern astronomical
        observations. Please consider double checking your time values
        and/or (format='{format}', scale='{scale}').
        """
        )

    return astropy_time


def set_times_from_astropy(self, astropy_time, is_barycentric=None):
    """
    Set the times for this Rainbow from an astropy Time object.

    Parameters
    ----------
    astropy_time : astropy.time.Time
        The times as an astropy Time object.
    is_barycentric : bool
        Are the times already measured relative to the
        Solar System barycenter? This is mostly for warning
        the user that it's not. Options are True, False,
        None (= don't know).


    Returns
    -------
    time : u.Quantity
        An astropy Quantity with units of time,
        expressing the Time as julian day.
        In addition to this returned variable,
        the function sets the following internal
        variables:

        self.time (= the astropy Quantity of times)
        self.metadata['time_format'] (= the format to convert back to Time)
        self.metadata['time_scale'] (= the scale to convert back to Time)
        self.metadata['time_is_barycentric'] (= is it barycentric?)
    """

    # set the formats
    format = "jd"
    unit = u.day
    scale = "tdb"

    # store the necessary values
    self.timelike["time"] = getattr(getattr(astropy_time, scale), format) * unit
    self.metadata["time_format"] = format
    self.metadata["time_scale"] = scale
    self.metadata["time_is_barycentric"] = is_barycentric

    # do some accounting to sync everything together
    self._guess_tscale()
    self._make_sure_time_edges_are_defined()
    return self.time



"""
Define a reader for chromatic .rainbow.npy files.
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_rainbow_npy"]


def from_rainbow_npy(rainbow, filepath):
    """
    Populate a Rainbow from a file in the .rainbow.npy format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the file to load, which should probably
        have an extension of `.rainbow.npy`
    """

    # read in your file, however you like
    loaded_core_dictionaries, version_used = np.load(filepath, allow_pickle=True)

    for k in rainbow._core_dictionaries:
        vars(rainbow)[k] = loaded_core_dictionaries[k]



"""
Define a reader for STScI pipeline x1dints.fits files.
"""
from ...imports import *

__all__ = ["from_coulombe"]


def from_coulombe(rainbow, filepath, order=1):
    """
    Populate a Rainbow from an STScI pipeline x1dints file,
    or a group of x1dints files for multiple segments.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file, or a glob-friendly string
        pointing to a group of files that should all be
        loaded together (for example, if an exposure was
        split into multiple segments).
    order : int
        The spectral order to load.
    """

    # create a list of filenames (which might be just 1)
    filenames = expand_filenames(filepath)

    # figure out the total number of integrations
    last_hdu = fits.open(filenames[-1])
    n_integrations_predicted_by_header = last_hdu["PRIMARY"].header["NINTS"]

    # populate a 1D array of times (with astropy units of time)
    fake_times = np.arange(n_integrations_predicted_by_header) * u.minute
    warnings.warn("The times are totally made up!")
    rainbow.timelike["time"] = fake_times

    # keep track of integrations loaded yet
    integration_counter = 0

    # loop over file (each one a segment)
    for i_file, f in enumerate(tqdm(filenames, leave=False)):

        # open this fits file
        hdu = fits.open(f)

        # there are three orders!
        n_integrations_in_segment = int((len(hdu) - 2) / 3)

        # if this is the first one, populate the shared stuff
        if i_file == 0:
            # grab the entire primary header
            rainbow.metadata["header"] = hdu["PRIMARY"].header

        start_extension = n_integrations_in_segment * (order - 1) + 1
        end_extension = start_extension + n_integrations_in_segment

        # loop through the spectra
        for e in range(start_extension, end_extension):

            # do this stuff only on the very first time through
            if i_file == 0 and e == start_extension:
                # pull out a wavelength grid (assuming it'll stay constant)
                wavelength_unit = u.Unit(hdu[e].columns["wavelength"].unit)
                rainbow.wavelike["wavelength"] = (
                    hdu[e].data["wavelength"] * wavelength_unit * 1
                )

                # set up the fluxlike quantities
                column_units = {}
                for column in hdu[e].columns:

                    # get a lower case name for the unit
                    c = column.name.lower()

                    # set up an empty Quantity array, with units if known
                    this_quantity = u.Quantity(
                        np.nan * np.ones((rainbow.nwave, rainbow.ntime))
                    )
                    if column.unit is not None:
                        column_units[c] = u.Unit(column.unit)
                        this_quantity *= column_units[c]
                    else:
                        column_units[c] = 1

                    # populate the fluxlike dictionary with the empty array
                    rainbow.fluxlike[c] = this_quantity * 1

            for column in hdu[e].columns:

                # get a lower case name for the unit
                c = column.name.lower()
                rainbow.fluxlike[c][:, integration_counter - 1] = (
                    hdu[e].data[c] * column_units[c] * 1
                )

            # increment the running integration total
            integration_counter += 1

    n_filled_times = np.sum(np.any(np.isfinite(rainbow.flux), rainbow.waveaxis))
    if n_filled_times != rainbow.ntime:
        warnings.warn(
            f"""
        The x1dints header(s) indicate there should be {rainbow.ntime} integrations,
        but only {n_filled_times} columns of the flux array were populated. Are you
        perhaps missing some segment files?
        """
        )

    # try to pull in the errors
    try:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["flux_error"] * 1
    except KeyError:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["error"] * 1
    other_order = {1: 2, 2: 1}[order]
    message = f"""
    Loading NIRISS order '{order}'. If you want the other order,
    trying `r = Rainbow(..., format='coulombe', order={other_order})`
    """
    warnings.warn(message)

    # try to guess wscale (and then kludge and call it linear)
    # rainbow._guess_wscale()
    # rainbow.metadata['wscale'] = 'linear' # TODO: fix this kludge



from ...imports import *

__all__ = ["inject_noise"]


def inject_noise(self, signal_to_noise=100, number_of_photons=None):
    """
    Inject uncorrelated random noise from a Gaussian
    or Poisson distribution into the flux array.

    Parameters
    ----------

    signal_to_noise : float, array
        The signal-to-noise per wavelength per time.
        For example, S/N=100 would mean that the
        uncertainty on the flux for each each
        wavelength-time data point will be 1%.
        If it is a scalar, then even point is the same.
        If it is an array with a fluxlike, wavelike,
        or timelike shape it will be broadcast
        appropriately.

    number_of_photons : float, array
        The number of photons expected to be recieved
        from the light source per wavelength and time.
        If it is a scalar, then even point is the same.
        If it is an array with a fluxlike, wavelike,
        or timelike shape it will be broadcast
        appropriately.

        If `number_of_photons` is set, then `signal_to_noise`
        will be ignored.

    Returns
    -------
    rainbow : Rainbow
        A new Rainbow object with the noise injected.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inject_noise", locals())

    # create a copy of the existing Rainbow
    new = self._create_copy()

    # get the underlying model (or create one if needed)
    if "model" in new.fluxlike:
        model = new.fluxlike["model"]
    else:
        # kludge, do we really want to allow this?
        model = self.flux * 1
        new.fluxlike["model"] = model

    # setting up an if/else statement so that the user
    # can choose if they want to use their own
    # number_of_photons or the automatic signal_to_noise
    # noise generation
    if number_of_photons is not None:
        if u.Quantity(model).unit != u.Unit(""):
            raise ValueError(
                f"""
            We haven't yet implemented `number_of_photons` noise
            for models that have units associated with them. Sorry!
            """
            )

        mu = model * self._broadcast_to_fluxlike(number_of_photons)

        # convert the model to photons and store it
        new.fluxlike["model"] = mu * u.photon

        # inject a realization of noise using number_of_photons
        # (yields poisson distribution)
        new.fluxlike["flux"] = np.random.poisson(mu) * u.photon  # mu is the center

        # store number of photons as metadata
        new.metadata["number_of_photons"] = number_of_photons

        # calculate the uncertainty
        uncertainty = np.sqrt(mu)
        new.fluxlike["uncertainty"] = uncertainty * u.photon

        # append the history entry to the new Rainbow
        new._record_history_entry(h)

    else:
        # calculate the uncertainty with a fixed S/N
        uncertainty = model / self._broadcast_to_fluxlike(signal_to_noise)
        new.fluxlike["uncertainty"] = uncertainty

        # inject a realization of the noise
        if isinstance(model, u.Quantity):
            unit = model.unit
            loc = model.to_value(unit)
            scale = uncertainty.to_value(unit)
        else:
            unit = 1
            loc = model
            scale = uncertainty
        new.fluxlike["flux"] = np.random.normal(model, uncertainty) * unit

        # store S/N as metadata
        new.metadata["signal_to_noise"] = signal_to_noise

        # append the history entry to the new Rainbow
        new._record_history_entry(h)

    # return the new object
    return new



"""
Define a writer for chromatic .rainbow.npy files.
"""

# import the general list of packages
from ...imports import *
from ...version import version

# define list of the only things that will show up in imports
__all__ = ["to_rainbow_npy"]


def to_rainbow_npy(self, filepath, **kw):
    """
    Write a Rainbow to a file in the .rainbow.npy format.

    Parameters
    ----------

    self : Rainbow
        The object to be saved.

    filepath : str
        The path to the file to write.
    """

    assert ".rainbow.npy" in filepath

    # populate a dictionary containing the four core dictionaries
    dictionary_to_save = self._get_core_dictionaries()

    # save that to a file
    np.save(filepath, [dictionary_to_save, version()], allow_pickle=True)



from .rainbow import *


class RainbowWithModel(Rainbow):
    """
    `RainbowWithModel` objects have a fluxlike `model`
    attached to them, meaning that they can

    This class definition inherits from `Rainbow`.
    """

    # which fluxlike keys will respond to math between objects
    _keys_that_respond_to_math = ["flux", "model"]

    # which keys get uncertainty weighting during binning
    _keys_that_get_uncertainty_weighting = ["flux", "model", "uncertainty"]

    @property
    def residuals(self):
        """
        Calculate the residuals on the fly,
        to make sure they're always up to date.
        """
        return self.flux - self.model

    @property
    def residuals_plus_one(self):
        """
        Calculate the residuals on the fly,
        to make sure they're always up to date.
        """
        return self.flux - self.model + 1

    @property
    def ones(self):
        """
        Generate an array of ones that looks like the flux.
        (A tiny wrapper needed for `plot_with_model`)
        """
        return np.ones_like(self.flux)

    def _validate_core_dictionaries(self):
        super()._validate_core_dictionaries()
        try:
            model = self.get("model")
            assert np.shape(model) == np.shape(self.flux)
        except (AttributeError, AssertionError):
            message = """
            No fluxlike 'model' was found attached to this
            `RainbowWithModel` object. The poor thing,
            its name is a lie! Please connect a model.
            The simplest way to do so might look like...
            `rainbow.model = np.ones(rainbow.shape)`
            ...or similarly with a more interesting array.
            """
            warnings.warn(message)

    from .visualizations import (
        plot_with_model,
        plot_with_model_and_residuals,
        imshow_with_models,
        plot_one_wavelength_with_models,
        animate_with_models,
    )


# REMOVE THE RAINBOW WITH MODEL AND JUST ADD A VALIDATION STEP TO ALL MODEL-DEPENDENT THINGS?



"""
Define a reader for STScI pipeline x1dints.fits files.
"""
from ...imports import *

__all__ = ["from_x1dints"]


def get_times_from_x1dints_files(filenames):
    """
    Wrapper to extract or estimate times from a list of file segments.

    Parameters
    ----------
    filenames : list
        The files to open, hopefully sorted.

    """
    try:
        timelike = {}
        for f in filenames:
            with fits.open(f) as hdu:
                table_of_times = hdu["int_times"].data
                for c in table_of_times.columns.names:
                    if c in timelike:
                        timelike[c].extend(list(table_of_times[c]))
                    else:
                        timelike[c] = list(table_of_times[c])

        for c in timelike:
            if "JD" in c:
                timelike[c] = timelike[c] * u.day
            else:
                timelike[c] = np.asarray(timelike[c])

        return timelike

    except KeyError:
        warnings.warn(
            f"""
        No `int_times` extension was found in the first file
        {filenames[0]})
        We're estimating the times from the `sci` extension
        using the TDB-BEG, TDB-END, and EFFINTTM keywords.
        """
        )

    try:
        with fits.open(filenames[0]) as hdu:

            # make up a time grid (this is still a little kludgy; it'd be better to have int_times)
            first_mjd_barycentric_integration_midpoint = (
                hdu["sci"].header["TDB-BEG"] * u.day
                + hdu["primary"].header["EFFINTTM"] / 2 * u.second
            )
            last_mjd_barycentric_integration_midpoint = (
                hdu["sci"].header["TDB-END"] * u.day
                - hdu["primary"].header["EFFINTTM"] / 2 * u.second
            )

        n_integrations_total = hdu["PRIMARY"].header["NINTS"]

        mjd_barycentric_integration_midpoints = np.linspace(
            first_mjd_barycentric_integration_midpoint,
            last_mjd_barycentric_integration_midpoint,
            n_integrations_total,
        )
        warnings.warn(
            f"""
        Times were set by linearly interpolating between the exposure
        start and end points. It's very possible these times are off
        by at least a few seconds and possibly up to the duration
        of one integration (= {hdu["primary"].header["EFFINTTM"]}s)
        """
        )
        return {"int_mid_BJD_TDB": mjd_barycentric_integration_midpoints}
    except KeyError:
        raise ValueError(
            f"""
        Blerg! We can't seem to determine the integration times
        from the `INT_TIMES` extension or even estimate them from
        the `SCI` extension.
        """
        )


def get_wavelengths_from_x1dints_files(
    filenames, order=1, non_spectrum_extensions=["PRIMARY", "INT_TIMES"]
):

    with fits.open(filenames[0]) as hdu:

        #
        e = (len(non_spectrum_extensions) - 1) + order - 1
        unit_string = hdu[e].columns["wavelength"].unit
        if unit_string is None:
            unit_string = "micron"
            warnings.warn("No wavelength unit was found; assuming 'micron'.")
        wavelength_unit = u.Unit(unit_string)
        return {"wavelength": hdu[e].data["wavelength"] * wavelength_unit * 1}


def from_x1dints(rainbow, filepath, order=1, **kw):
    """
    Populate a Rainbow from an STScI pipeline x1dints file,
    or a group of x1dints files for multiple segments.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated. (This is intended
        to be used as a class method of Rainbow or
        of a class derived from Rainbow, as a way of
        initializing an object from files.)
    filepath : str
        The path to the file, or a glob-friendly string
        pointing to a group of files that should all be
        loaded together (for example, if an exposure was
        split into multiple segments).
    """

    # create a list of filenames (which might be just 1)
    filenames = expand_filenames(filepath)

    # loop over file (each one a segment)
    for i_file, f in enumerate(filenames):

        # open this fits file
        with fits.open(f) as hdu:

            # if this is the first one, populate the shared stuff
            if i_file == 0:

                # grab the entire primary header
                rainbow.metadata["header"] = hdu["PRIMARY"].header

                if hdu["PRIMARY"].header["INSTRUME"] == "NIRISS":
                    n_orders = 3
                    warnings.warn(
                        f"""
                    Loading NIRISS spectroscopic `order={order}``. Three orders are available,
                    and you can set which (1,2,3) you want to read with the `order=` option.
                    """
                    )
                else:
                    n_orders = 1
                assert (order >= 1) and (order <= n_orders)

                non_spectrum_extensions = [
                    x for x in ["PRIMARY", "SCI", "INT_TIMES", "ASDF"] if x in hdu
                ]

            # set the index to the start of this segment
            integration_counter = hdu["PRIMARY"].header["INTSTART"]
            n_integrations_in_this_segment = (
                hdu["PRIMARY"].header["INTEND"] - hdu["PRIMARY"].header["INTSTART"] + 1
            )

            # make sure sizes match, ignoring PRIMARY, SCI, ASDF
            assert n_integrations_in_this_segment * n_orders == (
                len(hdu) - len(non_spectrum_extensions)
            )

            if i_file == 0:

                wavelike = get_wavelengths_from_x1dints_files(
                    filenames,
                    order=order,
                    non_spectrum_extensions=non_spectrum_extensions,
                )
                rainbow.wavelike.update(**wavelike)

                timelike = get_times_from_x1dints_files(filenames)
                rainbow.timelike.update(**timelike)
                astropy_times = Time(
                    timelike["int_mid_BJD_TDB"], format="mjd", scale="tdb"
                )
                rainbow.set_times_from_astropy(astropy_times, is_barycentric=True)

            # loop through the integrations in this segment
            for i in tqdm(np.arange(n_integrations_in_this_segment), leave=False):
                e = 2 + i * n_orders + (order - 1)

                # do this stuff only on the very first time through
                if i_file == 0 and i == 0:

                    # set up the fluxlike quantities
                    column_units = {}
                    for column in hdu[e].columns:

                        # get a lower case name for the unit
                        c = column.name.lower()

                        # set up an empty Quantity array, with units if known
                        this_quantity = u.Quantity(
                            np.nan * np.ones((rainbow.nwave, rainbow.ntime))
                        )
                        if column.unit is not None:
                            column_units[c] = u.Unit(column.unit)
                            this_quantity *= column_units[c]
                        else:
                            column_units[c] = 1

                        # populate the fluxlike dictionary with the empty array
                        rainbow.fluxlike[c] = this_quantity

                for column in hdu[e].columns:

                    # get a lower case name for the unit
                    c = column.name.lower()
                    rainbow.fluxlike[c][:, integration_counter - 1] = (
                        hdu[e].data[c] * column_units[c]
                    )

                # increment the running integration total
                integration_counter += 1

    n_filled_times = np.sum(np.any(np.isfinite(rainbow.flux), rainbow.waveaxis))
    if n_filled_times != rainbow.ntime:
        warnings.warn(
            f"""
        The x1dints header(s) indicate there should be {rainbow.ntime} integrations,
        but only {n_filled_times} columns of the flux array were populated. Are you
        perhaps missing some segment files?
        """
        )

    # try to pull in the errors
    try:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["flux_error"] * 1
    except KeyError:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["error"] * 1

    if rainbow.uncertainty is None:
        message = f"""
        Hmmm...it's not clear which column corresponds to the
        flux uncertainties for this Rainbow object. The
        available `fluxlike` columns are:
            {rainbow.fluxlike.keys()}
        A long-term solution might be to fix the `from_x1dints`
        reader, but a short-term solution would be to pick one
        of the columns listed above and say something like

        x.fluxlike['uncertainty'] = x.fluxlike['some-other-relevant-error-column']

        where `x` is the Rainbow you just created.
        """
        warnings.warn(message)



from ...imports import *

__all__ = ["remove_trends"]


def remove_trends(
    self, method="median_filter", **kw
):  # model=None,win_length=None,polyorder=None,butter_cutoff=None,butter_fs=None, butter_order=9,
    """
    A quick tool to approximately remove astrophysical signals,
    usually by some combination of smoothing/filtering.

    Parameters
    ----------
    method : str
        What method should be used to make an approximate model
        for smooth trends that will then be subtracted off?

        `differences` will do an extremely rough filtering
        of replacing the fluxes with their first differences.
        Trends that are smooth relative to the noise will
        be removed this way, but sharp features will remain.
        Required keywords:
            None.

        `median_filter` is a wrapper for scipy.signal.median_filter.
        It smoothes each data point to the median of its surrounding
        points in time and/or wavelength. Required keywords:
            `size` = centered on each point, what shape rectangle
            should be used to select surrounding points for median?
            The dimensions are (nwavelengths, ntimes), so `size=(3,7)`
            means we'll take the median across three wavelengths and
            seven times. Default is `(1,5)`.

        `savgol_filter` is a wrapper for scipy.signal.savgol_filter.
        It applies a Savitzky-Golay filter for polynomial smoothing.
        Required keywords:
            `window_length` = the length of the filter window,
            which must be a positive odd integer. Default is `5`.
            `polyorder` = the order of the polynomial to use.
            Default is `2`.

        `polyfit` is a wrapper for numpy.polyfit to use a weighted
        linear least squares polynomial fit to remove smooth trends
        in time. Required keywods:
            `deg` = the polynomial degree, which must be a positive
            integer. Default is `1`, meaning a line.

        `custom` allow users to pass any fluxlike array of model
        values for an astrophysical signal to remove it. Required
        keywords:
            `model` = the (nwavelengths, ntimes) model array

    kw : dict
        Any additional keywords will be passed to the function
        that does the filtering. See `method` keyword for options.

    Returns
    -------
    removed : Rainbow
        The Rainbow with estimated signals removed.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("remove_trends", locals())

    # TODO, think about more careful treatment of uncertainties + good/bad data
    new = self._create_copy()

    if method == "differences":
        new.flux = np.sqrt(2) * np.gradient(new.flux, axis=0) + 1

    #    if method == "butter_highpass":
    #        for i in range (0,new.nwave):
    #            nyq = 0.5 * butter_fs
    #            normal_cutoff = butter_cutoff/nyq
    #            b, a = butter(butter_order, normal_cutoff, btype = "high", analog = False)
    #            butter_filt = filtfilt(b, a, new.flux[i,:])
    #            new.flux[i,:] = new.flux[i,:]/butter_filt
    #
    #    if method == "convolve":
    #        for i in range (0,new.nwave):
    #            box = np.ones(win_length)/win_length
    #            grad = np.convolve(new.flux[i,:], box, mode = "same")
    #            new.flux[i,:] = new.flux[i,:]/grad

    if method == "median_filter":
        kw_to_use = dict(size=(1, 11))
        kw_to_use.update(**kw)
        if "size" not in kw:
            warnings.warn(
                f"""
            You didn't supply all expected keywords for '{method}'.
            Relying on defaults, the values will be:
            {kw_to_use}
            """
            )
        medfilt = median_filter(self.flux, **kw_to_use)
        new.flux = self.flux / medfilt
        new.uncertainty = self.uncertainty / medfilt

    if method == "savgol_filter":
        kw_to_use = dict(window_length=11, polyorder=1)
        kw_to_use.update(**kw)
        if ("window_length" not in kw) or ("polyorder" not in kw):
            warnings.warn(
                f"""
            You didn't supply all expected keywords for '{method}'.
            Relying on defaults, the values will be:
            {kw_to_use}
            """
            )
        for i in range(new.nwave):
            savgolfilter = savgol_filter(self.flux[i, :], **kw_to_use)
            new.flux[i, :] = self.flux[i, :] / savgolfilter
            new.uncertainty[i, :] = self.uncertainty[i, :] / savgolfilter

    if method == "polyfit":
        kw_to_use = dict(deg=1)
        kw_to_use.update(**kw)
        if "deg" not in kw:
            warnings.warn(
                f"""
            You didn't supply all expected keywords for '{method}'.
            Relying on defaults, the values will be:
            {kw_to_use}
            """
            )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for i in range(new.nwave):
                x, y, sigma = self.get_ok_data_for_wavelength(
                    i, express_badness_with_uncertainty=True
                )
                ok = np.isfinite(y)
                if np.sum(ok) >= 2:
                    try:
                        coefs = np.polyfit(
                            x=remove_unit(x)[ok],
                            y=remove_unit(y)[ok],
                            w=1 / remove_unit(sigma)[ok],
                            **kw_to_use,
                        )
                        poly = np.polyval(coefs, remove_unit(x))
                        new.flux[i, :] = self.flux[i, :] / poly
                        new.uncertainty[i, :] = self.uncertainty[i, :] / poly
                    except:
                        pass

    if method == "custom":
        if "model" not in kw:
            raise ValueError("You need a fluxlike `model` for this `custom` method")
        elif kw["model"].shape != new.flux.shape:
            raise ValueError("Your model doesn't match flux shape")
        else:
            new.flux = new.flux / kw["model"]
            new.uncertainty = new.uncertainty / kw["model"]

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new Rainbow
    return new



from .rainbow_npy import *
from .rainbow_FITS import *
from .xarray_stellar_spectra import *
from .xarray_raw_light_curves import *
from .xarray_fitted_light_curves import *

from .text import *


# construct a dictionary of available writers
available_writers = {k: globals()[k] for k in globals() if k[0:3] == "to_"}


def guess_writer(filepath, format=None):
    """
    A wrapper to guess the appropriate writer from the filename
    (and possibily an explicitly-set file format string).

    Parameters
    ----------
    filepath : str
        The path to the file to be written.
    format : str, None
        The file format to use.
    """
    from fnmatch import fnmatch
    import glob

    # get all the possible filenames (= expand wildcard)
    f = filepath

    # if format='abcdefgh', return the `to_abcdefgh` function
    if format is not None:
        return available_writers[f"to_{format}"]
    # does it look like a .rainbow.npy chromatic file?
    elif fnmatch(f, "*.rainbow.npy"):
        return to_rainbow_npy
    elif fnmatch(f.lower(), "*.rainbow.fits") or fnmatch(f.lower(), "*.rainbow.fit"):
        return to_rainbow_FITS
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*stellar-spec*.xc"):
        return to_xarray_stellar_spectra
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*raw-light-curve*.xc"):
        return to_xarray_raw_light_curves
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*fitted-light-curve*.xc"):
        return to_xarray_fitted_light_curves
    elif fnmatch(f, "*.txt") or fnmatch(filepath, "*.csv"):
        return to_text
    else:
        raise ValueError(
            f"""
        We're having trouble guessing the output format from the filename
        {f}
        Please try specifying a `format=` keyword to your `.save` call.
        """
        )



from ....imports import *

__all__ = ["imshow_quantities"]


def imshow_quantities(
    self, quantities=None, maxcol=3, panel_size=(5, 4), filename=None, **kw
):
    """
    imshow fluxlikes as a function of time (x = time, y = wavelength, color = flux).

    Parameters
    ----------
    quantities : list like
        The fluxlike quantity to imshow.
    maxcol : int
        The maximum number of columns to show (Optional).
    panel_size : tuple
        The (approximate) size of a single panel, which will
        be used to set the overall figsize based on the
        number of rows and columns (Optional).
    kw : dict
        Additional keywords will be passed on to `imshow`

    """

    # decide which quantities to plot
    if quantities is None:
        allkeys = self.fluxlike.keys()
    else:
        allkeys = quantities[:]

    # set up the geometry of the grid
    if len(allkeys) > maxcol:
        rows = int(np.ceil(len(allkeys) / maxcol))
        cols = maxcol
    else:
        rows = 1
        cols = np.min([len(allkeys), maxcol])

    # create the figure and grid of axes
    fig, axes = plt.subplots(
        rows,
        cols,
        figsize=(cols * panel_size[0], rows * panel_size[1]),
        sharex=True,
        sharey=True,
        constrained_layout=True,
    )

    # make the axes easier to index
    if len(allkeys) > 1:
        ax = axes.flatten()
    else:
        ax = [axes]

    # display each quantity
    for k, key in enumerate(allkeys):
        # make the imshow (or an empty box)
        if key in self.fluxlike.keys():
            self.imshow(quantity=key, ax=ax[k], **kw)
        else:
            ax[k].text(
                0.5,
                0.5,
                f"No {key}",
                transform=ax[k].transAxes,
                ha="center",
                va="center",
            )
        # add a title for each box
        ax[k].set_title(key)

        # hide xlabel except on the bottom row
        if k < (len(allkeys) - cols):
            ax[k].set_xlabel("")
        else:
            ax[k].tick_params(labelbottom=True)

        # hide ylabel except on the left column
        if (k % cols) > 0:
            ax[k].set_ylabel("")

    # hide any additional axes
    if k + 1 <= len(ax):
        for axi in ax[k + 1 :]:
            axi.axis("Off")
    if filename is not None:
        self.savefig(filename)



from ....imports import *

__all__ = [
    "get_for_wavelength",
    "get_ok_data_for_wavelength",
]


def get_for_wavelength(self, i, quantity="flux"):
    """
    Get 'quantity' associated with wavelength 'i'.

    Parameters
    ----------
    i : int
        The wavelength index to retrieve.
    quantity : string
        The quantity to retrieve. If it is flux-like,
        row 'i' will be returned. If it is time-like,
        the array itself will be returned.

    Returns
    -------
    quantity : np.array, u.Quantity
        The 1D array of 'quantity' corresponding to wavelength 'i'.
    """
    z = self.get(quantity)
    if np.shape(z) == self.shape:
        return z[i, :]
    elif len(z) == self.ntime:
        return z
    else:
        raise RuntimeError(
            f"""
        You tried to retrieve wavelength {i} from '{quantity}',
        but this quantity is neither flux-like nor time-like.
        It's not possible to return a time-like array. Sorry!
        """
        )


def get_ok_data_for_wavelength(
    self,
    i,
    x="time",
    y="flux",
    sigma="uncertainty",
    minimum_acceptable_ok=1,
    express_badness_with_uncertainty=False,
):
    """
    A small wrapper to get the good data from a wavelength,
    either trimming out data that are not OK or inflating the
    uncertainties to infinity.

    Parameters
    ----------
    i : int
        The wavelength index to retrieve.
    x : string
        What quantity should be retrieved as 'x'? (default = 'time')
    y : string
        What quantity should be retrieved as 'y'? (default = 'flux')
    sigma : string
        What quantity should be retrieved as 'sigma'? (default = 'uncertainty')
    minimum_acceptable_ok : float
        The smallest value of `ok` that will still be included.
        (1 for perfect data, 1e-10 for everything but terrible data, 0 for all data)
    express_badness_with_uncertainty : bool
        If False, data that don't pass the `ok` cut will be removed.
        If True, data that don't pass the `ok` cut will have their
        uncertainties inflated to infinity (np.inf).

    Returns
    -------
    x : np.array
        The time.
    y : np.array
        The desired quantity (default is `flux`)
    sigma : np.array
        The uncertainty on the desired quantity
    """

    # get 1D independent variable
    x_values = self.get_for_wavelength(i, x) * 1

    # get 1D array of what to keep
    ok = self.ok[i, :] >= minimum_acceptable_ok

    # get 1D array of the quantity
    y_values = self.get_for_wavelength(i, y) * 1

    # get 1D array of uncertainty
    sigma_values = self.get_for_wavelength(i, sigma) * 1

    if express_badness_with_uncertainty:
        sigma_values[ok == False] = np.inf
        return x_values, y_values, sigma_values
    else:
        return x_values[ok], y_values[ok], sigma_values[ok]



from ...imports import *
from ...resampling import leftright_to_edges

__all__ = ["pcolormesh"]


def pcolormesh(
    self,
    ax=None,
    quantity="flux",
    xaxis="time",
    w_unit="micron",
    t_unit="day",
    colorbar=True,
    mask_ok=True,
    color_ok="tomato",
    alpha_ok=0.8,
    vmin=None,
    vmax=None,
    filename=None,
    **kw,
):
    """
    Paint a 2D image of flux as a function of time and wavelength,
    using `plt.pcolormesh` where pixels can transform based on their edges.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    quantity : str
        The fluxlike quantity to imshow.
        (Must be a key of `rainbow.fluxlike`).
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    t_unit : str, astropy.unit.Unit
        The unit for plotting times.
    colorbar : bool
        Should we include a colorbar?
    mask_ok : bool
        Should we mark which data are not OK?
    color_ok : str
        The color to be used for masking data points that are not OK.
    alpha_ok : float
        The transparency to be used for masking data points that are not OK.
    kw : dict
        All other keywords will be passed on to `plt.pcolormesh`,
        so you can have more detailed control over the plot
        appearance. Common keyword argumentsvli might include:
        `[cmap, norm, alpha, vmin, vmax]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.pcolormesh.html
    """

    # self.speak(f'imshowing')
    if ax is None:
        ax = plt.subplot()

    # get units
    w_unit, t_unit = u.Unit(w_unit), u.Unit(t_unit)

    # make sure some wavelength and time edges are defined
    self._make_sure_wavelength_edges_are_defined()
    self._make_sure_time_edges_are_defined()

    # set up the wavelength and time edges
    w_edges = leftright_to_edges(
        self.wavelength_lower.to_value(w_unit), self.wavelength_upper.to_value(w_unit)
    )
    t_edges = leftright_to_edges(
        self.time_lower.to_value(t_unit), self.time_upper.to_value(t_unit)
    )

    wlabel = f"{self._wave_label} ({w_unit.to_string('latex_inline')})"
    tlabel = f"{self._time_label} ({t_unit.to_string('latex_inline')})"

    def get_2D(k):
        """
        A small helper to get a 2D quantity. This is a bit of
        a kludge to help with weird cases of duplicate keys
        (for example where 'wavelength' might appear in both
        `wavelike` and `fluxlike`).
        """
        z = self.get(k)
        if np.shape(z) == self.shape:
            return z
        else:
            return self.fluxlike.get(k, None)

    if xaxis.lower()[0] == "t":
        x, y = t_edges, w_edges
        xlabel, ylabel = tlabel, wlabel
        z = get_2D(quantity)
        ok = get_2D("ok")
    elif xaxis.lower()[0] == "w":
        x, y = w_edges, t_edges
        xlabel, ylabel = wlabel, tlabel
        z = get_2D(quantity).T
        ok = get_2D("ok").T
    else:
        warnings.warn(
            "Please specify either `xaxis='time'` or `xaxis='wavelength'` for `.plot()`"
        )

    # figure out a good shared color limits (unless already supplied)
    vmin = vmin or np.nanpercentile(u.Quantity(z.flatten()).value, 1)
    vmax = vmax or np.nanpercentile(u.Quantity(z.flatten()).value, 99)

    # define some default keywords
    pcolormesh_kw = dict(shading="flat", vmin=vmin, vmax=vmax)
    pcolormesh_kw.update(**kw)
    with quantity_support():
        plt.sca(ax)
        if mask_ok:
            okpcolormesh_kw = dict(**pcolormesh_kw)
            okpcolormesh_kw.update(
                cmap=one2another(
                    bottom=color_ok,
                    top=color_ok,
                    alpha_bottom=alpha_ok,
                    alpha_top=0,
                ),
                zorder=10,
                vmin=0,
                vmax=1,
            )
            plt.pcolormesh(
                x,
                y,
                ok,
                **okpcolormesh_kw,
            )
        plt.pcolormesh(
            x,
            y,
            z,
            **pcolormesh_kw,
        )
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        if colorbar:
            plt.colorbar(
                ax=ax,
                label=u.Quantity(z).unit.to_string("latex_inline"),
            )
        # emulate origin = upper for imshow (y starts at top)
        plt.ylim(y[-1], y[0])
        plt.title(self.get("title"))

    if filename is not None:
        self.savefig(filename)
    return ax



from ....imports import *

__all__ = ["get_expected_uncertainty"]


def get_expected_uncertainty(self, function=np.nanmedian, *args, **kwargs):
    """
    Get the typical per-wavelength uncertainty.

    Parameters
    ----------
    function : function
        What function should be used to choose the "typical"
        value for each wavelength? Good options are probably
        things like `np.nanmedian`, `np.median`, `np.nanmean`
        `np.mean`, `np.percentile`
    args : list
        Addition arguments will be passed to `function`
    kw : dict
        Additional keyword arguments will be passed to `function`

    Returns
    -------
    uncertainty_per_wavelength : np.array (wavelike)
        The uncertainty associated with each wavelength.
    """
    uncertainty_per_wavelength = function(
        self.uncertainty, *args, axis=self.timeaxis, **kwargs
    )
    return uncertainty_per_wavelength



from ..rainbows import *

closekw = dict(rtol=1e-10)


def test_to_df():

    r = SimulatedRainbow(dt=1 * u.minute, R=50).inject_noise(signal_to_noise=10)

    r_df = r.to_df()

    # ensure the length of the df is the same as the original rainbow
    assert len(r_df) == r.nflux

    # ensure we have the right column names
    columnnames = r_df.columns
    for colname in ["Time (d)", "Wavelength (micron)", "Flux", "Flux Uncertainty"]:
        assert colname in columnnames

    # check the values in the df match the rainbow
    assert np.isclose(
        r_df["Time (d)"].values[0], r.time.to_value("h")[0] / 24.0, **closekw
    )  # default=days
    assert r_df["Wavelength (micron)"].values[0] == r.wavelength.to_value("micron")[0]
    assert r_df["Flux"].values[0] == r.flux[0, 0]
    assert r_df["Flux Uncertainty"].values[0] == r.uncertainty[0, 0]

    # test the timeformat parameter
    for t_unit in ["h", "hour", "day", "minute", "second", "s"]:
        r_df = r.to_df(t_unit=t_unit)
        assert f"Time ({t_unit})" in r_df.columns


def test_to_nparray():
    r = SimulatedRainbow(dt=1 * u.minute, R=50).inject_noise(signal_to_noise=100)

    rflux, rfluxu, rtime, rwavel = r.to_nparray()

    assert len(rtime) == r.ntime
    assert len(rwavel) == r.nwave
    assert len(rtime) * len(rwavel) == r.nflux
    assert len(rflux.flatten()) == r.nflux
    assert len(rfluxu.flatten()) == r.nflux
    assert np.shape(rflux) == r.shape
    assert np.shape(rfluxu) == r.shape

    assert np.all(rflux == r.flux)
    assert np.all(rfluxu == r.uncertainty)
    assert np.all(rwavel == r.wavelength.to_value("micron"))

    # issues with rounding errors:
    assert np.all(
        np.isclose(rtime, r.time.to_value("h") / 24.0, **closekw)
    )  # the default is days

    # test if the hours format works
    rflux, rfluxu, rtime, rwavel = r.to_nparray(t_unit="h")
    assert np.all(np.isclose(rtime, r.time.to_value("h"), **closekw))

    # test if the minutes format works
    rflux, rfluxu, rtime, rwavel = r.to_nparray(t_unit="min")
    assert np.all(np.isclose(rtime, r.time.to_value("h") * 60, **closekw))

    # test if the minutes format works
    rflux, rfluxu, rtime, rwavel = r.to_nparray(t_unit="s")
    assert np.all(np.isclose(rtime, r.time.to_value("h") * 3600, **closekw))



from ..rainbows import *
from .setup_tests import *


def test_astropy_times():
    s = SimulatedRainbow()
    format = "jd"
    scale = "tdb"
    original_times = s.time
    astropy_times = s.get_times_as_astropy(format=format, scale=scale)
    s.set_times_from_astropy(astropy_times)

    assert np.all(s.time == original_times)
    assert np.all(astropy_times == s.get_times_as_astropy())



from ..rainbows import *
from .setup_tests import *


def test_remove_trends():
    s = (
        SimulatedRainbow(dw=0.1 * u.micron)
        .inject_transit()
        .inject_systematics()
        .inject_noise(signal_to_noise=300)
    )
    fi, ax = plt.subplots(
        2,
        6,
        figsize=(12, 6),
        sharey="row",
        sharex=True,
        dpi=300,
        constrained_layout=True,
    )
    imkw = dict(vmin=0.98, vmax=1.02, xaxis="wavelength", colorbar=False)
    s.imshow(ax=ax[0, 0], **imkw)
    s.plot_noise_comparison(ax=ax[1, 0])
    for i, method in enumerate(
        ["median_filter", "savgol_filter", "differences", "polyfit"]
    ):
        x = s.remove_trends(method=method)
        x.imshow(ax=ax[0, 1 + i], **imkw)
        ax[0, i + 1].set_title(method)
        x.plot_noise_comparison(ax=ax[1, i + 1])
    x = s.remove_trends(method="custom", model=s.planet_model)
    ax[0, -1].set_title("custom")

    x.imshow(ax=ax[0, -1], **imkw)
    x.plot_noise_comparison(ax=ax[1, -1])
    plt.ylim(0, 0.02)
    plt.savefig(os.path.join(test_directory, "test-remove_trends.png"))



from ..rainbows import *
from .setup_tests import *


def test_inject_systematics():
    SimulatedRainbow().inject_transit().inject_noise().inject_systematics().bin(
        R=10, dt=10 * u.minute
    ).imshow_quantities()
    plt.savefig(os.path.join(test_directory, "inject_systematics_demo.pdf"))



from ....imports import *

__all__ = ["get_measured_scatter"]


def get_measured_scatter(
    self, quantity="flux", method="standard-deviation", minimum_acceptable_ok=1e-10
):
    """
    Get measured scatter from a chromatic rainbow.

    Parameters
    ----------
    quantity : string
        The fluxlike quantity for which we should calculate the scatter.
    method : string
        What method to use to obtain measured scatter. Current options are 'MAD', 'standard-deviation'.
    minimum_acceptable_ok : float
        The smallest value of `ok` that will still be included.
        (1 for perfect data, 1e-10 for everything but terrible data, 0 for all data)


    Returns
    -------
    scatter : np.array (wavelike)
        The scatter wavelike array.
    """

    if method not in ["standard-deviation", "MAD"]:
        warnings.warn(
            f"""
        '{method}' is not an available method.
        Please choose from ['MAD', 'standard-deviation'].
        """
        )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        scatters = np.zeros(self.nwave)
        for i in range(self.nwave):
            x, y, sigma = self.get_ok_data_for_wavelength(
                i, y=quantity, minimum_acceptable_ok=minimum_acceptable_ok
            )
            if u.Quantity(y).unit == u.Unit(""):
                y_value, y_unit = y, 1
            else:
                y_value, y_unit = y.value, y.unit
            if method == "standard-deviation":
                scatters[i] = np.nanstd(y_value)
            elif method == "MAD":
                scatters[i] = median_absolute_deviation(y_value, scale="normal")
        return scatters * y_unit



from ...imports import *

# don't make Altair a necessary part of chromatic
try:
    import altair as alt

    alt.data_transformers.disable_max_rows()
except Exception as e:
    print(e)
    warnings.warn(
        "Issue importing Altair, cannot make interactive plot :(! \n \
                  You can install Altair using: pip install altair"
    )

__all__ = ["imshow_interact"]


# Convert this grid to columnar data expected by Altair
def imshow_interact(
    self,
    quantity="Flux",
    t_unit="d",
    w_unit="micron",
    cmap="viridis",
    ylim=[],
    ylog=None,
    filename=None,
):
    """
    Display interactive spectrum plot for chromatic Rainbow with a
    wavelength-averaged 2D quantity defined by the user. The user
    can interact with the 3D spectrum to choose the wavelength range
    over which the average is calculated.

    Parameters
    ----------
    self : Rainbow object
        chromatic Rainbow object to plot
    quantity : str
        (optional, default='flux')
        The quantity to imshow, currently either `flux` or `uncertainty`
    ylog : boolean
        (optional, default=None)
        Boolean for whether to take log10 of the y-axis data.
        If None, will be guessed from the data.
    t_unit : str
        (optional, default='d')
        The time unit to use (seconds, minutes, hours, days etc.)
    w_unit : str
        (optional, default='micron')
        The wavelength unit to use
    cmap : str
        (optional, default='viridis')
        The color scheme to use from Vega documentation
    ylim : list
        (optional, default=[])
        If the user wants to define their own ylimits on the lightcurve plot
    """

    # preset the x and y axes as Time (in units defined by the user) and Wavelength
    xlabel = f"Time ({t_unit})"
    ylabel = f"Wavelength ({w_unit})"

    # allow the user to plot flux or uncertainty
    if quantity.lower() == "flux":
        z = "Flux"
    elif quantity.lower() == "uncertainty":
        z = "Flux Uncertainty"
    elif quantity.lower() == "error":
        z = "Flux Uncertainty"
    elif quantity.lower() == "flux_error":
        z = "Flux Uncertainty"
    elif quantity.lower() == "flux_uncertainty":
        z = "Flux Uncertainty"
    else:
        # if the quantity is not one of the predefined values:
        warnings.warn("Unrecognised Quantity!")
        return

    # convert rainbow object to pandas dataframe
    source = self.to_df(t_unit=t_unit, w_unit=w_unit)[[xlabel, ylabel, z]]

    # if there are >10,000 data points Altair will be very laggy/slow. This is probably unbinned, therefore
    # encourage the user to bin the Rainbow before calling this function in future/
    N_warning = 100000
    if len(source) > N_warning:
        warnings.warn(
            f"""
        The dataset {self} has {N_warning} data points/
        The interactive plot may lag. Try binning first!
        """
        )

    if self._is_probably_normalized() == False:
        warnings.warn(
            """
        It looks like you might be trying to use `imshow_interact` with an
        unnormalized Rainbow object. You might consider normalizing first
        with `rainbow.normalize().imshow_interact()`.
        """
        )

    # The unbinned Rainbow is sometimes in log scale, therefore plotting will be ugly with uniform axis spacing
    # ylog tells the function to take the log10 of the y-axis data
    try:
        ylog = ylog or (self.wscale == "log")
    except AttributeError:
        ylog = ylog or False

    if ylog:
        source[ylabel] = np.log10(source[ylabel])
        source = source.rename(columns={ylabel: f"log10({ylabel})"})
        ylabel = f"log10({ylabel})"

    if len(ylim) > 0:
        domain = ylim
    else:
        domain = [source[z].min() - 0.005, source[z].max() + 0.005]

    # Add interactive part
    brush = alt.selection(type="interval", encodings=["y"])

    # Define the 3D spectrum plot
    spectrum = (
        alt.Chart(source, width=280, height=230)
        .mark_rect(
            clip=False,
            width=280 / len(self.timelike["time"]),
            height=230 / len(self.wavelike["wavelength"]),
        )
        .encode(
            x=alt.X(
                f"{xlabel}:Q",
                scale=alt.Scale(
                    zero=False,
                    nice=False,
                    domain=[np.min(source[xlabel]), np.max(source[xlabel])],
                ),
            ),
            y=alt.Y(
                f"{ylabel}:Q",
                scale=alt.Scale(
                    zero=False,
                    nice=False,
                    domain=[np.max(source[ylabel]), np.min(source[ylabel])],
                ),
            ),
            fill=alt.Color(
                f"{z}:Q",
                scale=alt.Scale(
                    scheme=cmap,
                    zero=False,
                    domain=domain,
                ),
            ),
            tooltip=[f"{xlabel}", f"{ylabel}", f"{z}"],
        )
    )

    # gray out the background with selection
    background = spectrum.encode(color=alt.value("#ddd")).add_selection(brush)

    # highlights on the transformed data
    highlight = spectrum.transform_filter(brush)

    # Layer the various plotting parts
    spectrum_int = alt.layer(background, highlight, data=source)

    # Add the 2D averaged lightcurve (or uncertainty)
    lightcurve = (
        alt.Chart(source, width=280, height=230, title=f"Mean {z} for Wavelength Range")
        .mark_point(filled=True, size=20, color="black")
        .encode(
            x=alt.X(
                f"{xlabel}:Q",
                scale=alt.Scale(
                    zero=False,
                    nice=False,
                    domain=[
                        np.min(source[xlabel])
                        - (0.02 * np.abs(np.min(source[xlabel]))),
                        1.02 * np.max(source[xlabel]),
                    ],
                ),
            ),
            y=alt.Y(
                f"mean({z}):Q",
                scale=alt.Scale(zero=False, domain=domain),
                title="Mean " + z,
            ),
        )
        .transform_filter(brush)
    )

    # display the interactive Altair plot
    (spectrum_int | lightcurve).display()
    if filename is not None:
        (spectrum_int | lightcurve).save(self._label_plot_file(filename))



from ..imports import *
from ..resampling import bintoR, bintogrid
from ..version import __version__
import astropy.config.paths
from astropy.utils.data import is_url_in_cache, cache_total_size


class PHOENIXLibrary:
    # downloaded files will be stored in "~/.{_cache_label}"
    _cache_label = "chromatic"

    # these coordinates are needed to index the grid (in this order)
    _keys_for_indexing = ["temperature", "logg", "metallicity"]

    # the resolutions available as precalculated grids
    _available_resolutions = [
        3,
        10,
        30,
        100,
        300,
        1000,
        3000,
        10000,
        30000,
        100000,
    ]

    def get_cache_dir(self):
        return astropy.config.paths.get_cache_dir(self._cache_label)

    def get_cache_size(self):
        return (cache_total_size(self._cache_label) * u.byte).to(u.gigabyte)

    def __init__(self, directory=".", photons=True):
        """
        Initialize a PHOENIX model library to provide easy
        access to model stellar spectra at resolutions up
        to R=100000.

        Parameters
        ----------
        directory : str
            The path to a directory where *new* library grids
            will be created. This will only be used if you're
            calling the super secret `._create_grids` functions
            to download raw PHOENIX spectrum files and compile
            them into a new library. Most users can just
            not worry about this!
        photons : bool
            Should the units be in photons, rather than power?
            If True, spectrum units will be photons/(s * m**2 * nm)
            If False, spectrum units will be J/(s * m**2 * nm)
        """

        # should we use photon units (e.g. photons/s/m**2/nm)?
        # (if not, we'll use flux per wavelength units like W/m**2/nm)
        self._are_the_units_photons = photons

        # figure out the directory where the models will be stored
        self._directory_for_new_grids = os.path.join(
            os.path.join(directory),
            "chromatic-model-stellar-spectra",
        )

        self._local_paths = {}

    def _download_grid(self, R, metallicity=0.0, cache=True):
        """
        Download the preprocessed grid for a particular
        resolution, or load it from a cached local file.

        Parameters
        ----------
        R : float
            The resolution of the grid to retrieve
        metallicity : float
            The stellar metallicity (= log10[metals/solar])
        cache : bool
            Once it's downloaded, should we keep it for next time?

        Returns
        -------
        filename : str
            The path to the downloaded local file for the grid.
        """

        # make sure the resolution is reasonable
        if R not in self._available_resolutions:
            raise ValueError(
                f"""
            Your requested resolution of R={R} is not in
            {self._available_resolutions}
            Please pick one of those choices to download.
            """
            )

        # make sure it's a possible R
        assert R in self._available_resolutions

        # download (or find the cached local file)
        basename = self._get_grid_filename(R, metallicity=metallicity)
        url = f"https://casa.colorado.edu/~bertathompson/chromatic/{basename}"

        if is_url_in_cache(url, pkgname=self._cache_label):
            self._local_paths[R] = download_file(
                url, pkgname=self._cache_label, cache=True
            )
        else:

            threshold = 1000
            if R <= threshold:
                expected_time = f"Because the resolution is R<={threshold}, this should be pretty quick."
            else:
                expected_time = f"Because the resolution is R>{threshold}, this might be annoyingly slow."
            print(
                f"""
            Downloading pre-processed grid for R={R}, metallicity={metallicity} from
            {url}
            {expected_time}
            """
            )

            self._local_paths[R] = download_file_with_warning(
                url, pkgname=self._cache_label, cache=cache, show_progress=True
            )
        return self._local_paths[R]

    def _download_raw_data(self, metallicity=0.0, cache=True):
        """
        Make sure the raw data from the online PHOENIX database
        are downloaded to your local computer. (Most users
        shouldn't have to interact with this, unless you
        want to do something particularly fancy.)

        You must be connected to the internet for this to work.
        It may take quite a long time!

        Parameters
        ----------
        metallicity : float
            The stellar metallicity (= log10[metals/solar])
        cache : bool
            Once it's downloaded, should we keep it for next time?
        """

        # create a dictionary to store the local
        self._raw_local_paths = {}

        # where are the raw data located at online?
        self._raw_base_url = "ftp://phoenix.astro.physik.uni-goettingen.de/HiResFITS"

        # get the one shared wavelength array
        self._raw_wavelengths_filename = "WAVE_PHOENIX-ACES-AGSS-COND-2011.fits"
        self._raw_wavelengths_url = "/".join(
            [self._raw_base_url, self._raw_wavelengths_filename]
        )
        warnings.warn(
            f"""
        Downloading (or finding locally) the shared wavelength grid from
        {self._raw_wavelengths_url}
        """
        )
        self._raw_local_paths["wavelengths"] = download_file_with_warning(
            self._raw_wavelengths_url, pkgname=self._cache_label, cache=cache
        )

        # get the index of files in this metallicity's directory
        metallicity_string = self._stringify_metallicity(metallicity)
        self._raw_directory = f"PHOENIX-ACES-AGSS-COND-2011/Z{metallicity_string}"
        self._raw_directory_url = "/".join([self._raw_base_url, self._raw_directory])
        self._raw_local_paths[
            f"index-Z={metallicity_string}"
        ] = download_file_with_warning(
            self._raw_directory_url, pkgname=self._cache_label, cache=cache
        )
        warnings.warn(
            f"""
        Downloading (or finding locally) the index of files from
        {self._raw_directory_url}
        """
        )

        # get the complete list of spectrum files
        self._raw_spectrum_filenames = list(
            [
                x
                for x in ascii.read(
                    self._raw_local_paths[f"index-Z={metallicity_string}"]
                )
                .columns[-1]
                .data
                if ".fits" in x
            ]
        )
        self._raw_spectrum_urls = [
            "/".join([self._raw_base_url, self._raw_directory, f])
            for f in self._raw_spectrum_filenames
        ]
        N = len(self._raw_spectrum_urls)
        warnings.warn(
            f"""
        Downloading (or finding locally) {N} very large files from
        {self._raw_directory_url}

        If the files aren't already downloaded,
        this might take an annoyingly long time!
        If it crashes due to a timeout,
        try restarting.
        """
        )
        self._current_raw_metallicity = metallicity
        self._raw_downloaded = {}
        for url, file in tqdm(
            zip(self._raw_spectrum_urls, self._raw_spectrum_filenames), leave=False
        ):
            self._raw_downloaded[file] = download_file_with_warning(
                url, pkgname=self._cache_label, cache=cache
            )

    def _load_raw_wavelength(self):
        """
        Load in the raw wavelength array.

        Returns
        -------
        wavelength : u.Quantity
            The wavelengths associated with this grid,
            with astropy units of microns.
        """

        # default to the preloaded raw wavelength
        try:
            return self._raw_wavelength

        # load the raw wavelengths and save them for next time
        except AttributeError:
            wavelength_filename = self._raw_local_paths["wavelengths"]
            hdu = fits.open(wavelength_filename)
            wavelength_without_unit = hdu[0].data
            wavelength_unit = u.Angstrom
            wavelength = wavelength_without_unit * wavelength_unit
            self._raw_wavelength = wavelength.to("micron")
            return self._raw_wavelength

    def _load_raw_spectrum(self, filename):
        """
        Load in a raw spectrum array.

        Parameters
        ----------
        filename : string
            The filename of the raw PHOENIX spectrum
        Returns
        -------
        spectrum : u.Quantity
            The spectrum, with astropy units of W/(m**2 nm)
        """
        hdus = fits.open(filename)
        flux_without_unit = hdus[0].data
        flux_unit = u.Unit("erg/(s * cm**2 * cm)")
        flux = flux_without_unit * flux_unit
        return flux.to("W/(m**2 nm)")

    def _stringify_metallicity(self, metallicity):
        """
        Convert a metallicity into a PHOENIX-style string.

        Parameters
        ----------
        metallicity : float
            [Fe/H]-style metallicity (= 0.0 for solar)

        Returns
        -------
        s : string
            The metallicity, as a PHOENIX-style string.
        """
        if metallicity <= 0:
            return f"-{np.abs(metallicity):03.1f}"
        else:
            return f"+{metallicity:03.1f}"

    def _get_Tgz_from_filename(self, filename):
        """
        A helper to get the temperature, logg, and metallicity
        from a PHOENIX spectrum model filename.

        Parameters
        ----------
        filename : string
            The filename of PHOENIX model.

        Returns
        -------
        temperature : float
            Temperature (K)
        logg : float
            log10([surface gravity]/[cm/s**2])
        metallicity : float
            log10([metallicity]/[solar metallicity])
        """

        f = os.path.basename(filename)
        temperature = float(f[3:8])
        logg = float(f[9:13])
        metallicity = float(f[13:17].replace("-0.0", "0.0"))
        return temperature, logg, metallicity

    def _get_filename_from_Tgz(self, temperature, logg, metallicity):
        """
        A helper to get a PHOENIX spectrum model filename
        from temperature, logg, and metallicity.

        Parameters
        ----------
        T : float
            Temperature (K)
        logg : float
            log10([surface gravity]/[cm/s**2])
        metallicity : float
            log10([metallicity]/[solar metallicity])

        Returns
        -------
        filename : string
            The filename of PHOENIX model (excluding directory).
        """

        f"lte{temperature:05.0f}-{logg:04.2f}{self._stringify_metallicity(metallicity)}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"
        return f

    def _get_grid_filename(self, R, metallicity=0.0):
        """
        Get the filename of a pre-processed grid.

        Parameters
        ----------
        R : float
            The resolution of the grid.
        metallicity : float
            The stellar metallicity.

        Returns
        -------
        filename : str
            The filepath to the pre-processed grid.
        """
        if self._are_the_units_photons:
            unit_string = "photons"
        else:
            unit_string = "flux"

        filename = f"phoenix_{unit_string}_metallicity={metallicity:3.1f}_R={R:.0f}.npy"
        return filename

    def _create_grids(self, remake=False):
        """
        Create pre-processed grids for all resolutions.

        Parameters
        ----------
        metallicity : float
            The stellar metallicity.
        remake : bool
            Should we remake the library even if a file exists?
        """
        for metallicity in self._available_metallicities:
            for R in self._available_resolutions:
                self._create_grid(
                    R,
                    metallicity=metallicity,
                    remake=remake,
                )

    def _create_upload_command_for_grids(self, destination):
        """
        A tiny wrapper to generate a UNIX command that
        can upload locally-generated library files to
        an online source.

        Returns
        -------
        command : string
            A copy-paste `rsync` command to upload new files.
        """
        source = os.path.join(self._directory_for_new_grids, "phoenix_*.npy")
        command = f"rsync -v --progress {source} {destination}"
        return command

    def _create_grid(self, R, metallicity=0.0, remake=False):
        """
        Create a pre-processed grid for a single resolution.

        Parameters
        ----------
        R : float
            The resolution of the grid.
        metallicity : float
            The stellar metallicity.
        remake : bool
            Should we remake the library even if a file exists?
        """

        # make sure that directory exists
        try:
            os.mkdir(self._directory_for_new_grids)
        except FileExistsError:
            pass

        try:
            assert self._current_raw_metallicity == metallicity
            assert len(self._raw_spectrum_filenames) == len(self._raw_downloaded)
        except (AttributeError, AssertionError):
            self._download_raw_data(metallicity=metallicity)

        # skip this resolution if already made
        filename = os.path.join(
            self._directory_for_new_grids,
            self._get_grid_filename(R, metallicity=metallicity),
        )
        if os.path.exists(filename) and (not remake):
            print(
                textwrap.dedent(
                    f"""
                a grid for R={R}, metallicity={metallicity} exists at
                {filename}
                so we're not remaking it
                """
                )
            )
            return

        print(
            f"Creating a new grid for R={R}, metallicity={metallicity}. Its details are..."
        )
        shared = {}
        shared["grid"] = "PHOENIX-ACES-AGSS-COND-2011"
        shared["url"] = "https://phoenix.astro.physik.uni-goettingen.de/?page_id=15"
        shared["citation"] = "2013A&A…553A…6H"
        shared["photons"] = self._are_the_units_photons
        shared["R"] = R
        shared["metallicity"] = metallicity
        shared["filename"] = os.path.basename(filename)
        shared["chromatic-version"] = __version__
        unbinned_w = self._load_raw_wavelength()
        for k, v in shared.items():
            print(f"{k:>20} = {v}")

        d = {}
        for k, v in tqdm(list(self._raw_downloaded.items()), leave=False):

            # load the unbinned spectrum
            unbinned_f = self._load_raw_spectrum(v)

            # convert from W to photon/s
            if self._are_the_units_photons:
                photon_energy = (con.h * con.c / unbinned_w) / u.photon
                unbinned_f = (unbinned_f / photon_energy).to("ph/(s m**2 nm)")
            else:
                unbinned_f = unbinned_f.to("W/(m**2 nm)")
            # figure out its stellar inputs
            T, logg, Z = self._get_Tgz_from_filename(k)
            if R == "original":
                w = unbinned_w
                f = unbinned_f
            else:
                binned = bintoR(unbinned_w, unbinned_f, R=R, drop_nans=False)
                w, f = binned["x"], binned["y"]

            if "wavelength" not in shared:
                shared["wavelength"] = w.value

            assert len(f) == len(shared["wavelength"])

            key = (T, logg, Z)
            d[key] = f.value

        # pull out the unique values of the keys
        for i, k in enumerate(self._keys_for_indexing):
            shared[k] = np.unique([x[i] for x in d])

        shared["wavelength_unit"] = w.unit.to_string()
        shared["spectrum_unit"] = f.unit.to_string()

        # save everything to an easy-to-load file
        np.save(filename, [shared, d], allow_pickle=True)
        print(f"That grid has been saved to {filename}.\n")

    def _find_smallest_R(self, R):
        """
        Make sure the smallest grid necessary to support
        a particular resolution is loaded into the library.

        Parameters
        ----------
        R : float
            The resolution of the grid.
        """

        ok = np.array(self._available_resolutions) >= R
        if np.any(ok):
            i = np.flatnonzero(ok)[0]
        else:
            warnings.warn(
                f"""
            Your request resolution of R={R} is higher than the maximum.
            It will be rounded down to R={np.max(self._available_resolutions)}.
            """
            )
            i = -1

        if i is not None:
            smallest_sufficient_R = self._available_resolutions[i]
        else:
            warnings.warn(
                f"""
            Your requested resolution of R={R}
            is higher than the largest possible value
            with uniform binning (R={np.max(self._available_resolutions)}).
            """
            )
            smallest_sufficient_R = np.max(self._available_resolutions)
        return smallest_sufficient_R

    def _get_local_grid(self, R, metallicity=0.0, directory="."):
        bespoke = os.path.join(
            self._directory_for_new_grids,
            self._get_grid_filename(R, metallicity=metallicity),
        )
        if os.path.exists(bespoke):
            return bespoke
        else:
            return self._download_grid(R, metallicity=metallicity)

    def _load_grid(self, R, metallicity=0.0):
        """
        Load a pre-processed grid for a particular resolution.

        Parameters
        ----------
        R : float
            The resolution of the grid.
        metallicity : float
            The stellar metallicity.
        """
        try:
            assert self.metadata["R"] == R
            assert metallicity in self.metadata["metallicity"]
            return
        except (AttributeError, AssertionError):
            pass

        metadata, models = np.load(
            self._get_local_grid(R, metallicity=metallicity), allow_pickle=True
        )[()]

        try:
            for k in ["R", "photons", "wavelength"]:
                assert np.all(self.metadata[k] == metadata[k])
            for k in ["metallicity", "filename", "chromatic-version"]:
                self.metadata[k] = np.hstack([self.metadata[k], metadata[k]])
            self.models.update(**models)

        except (AttributeError, AssertionError):
            self.metadata = metadata
            self.models = models
            self.wavelength_cached_models = {}

        self.units = {
            k: u.Unit(self.metadata[f"{k}_unit"]) for k in ["wavelength", "spectrum"]
        }
        self.wavelength = self.metadata["wavelength"] * self.units["wavelength"]

    def _find_bounds(self, value, key, inputs={}):
        """
        Find a mask where `possible` is immediately
        below or above `value`.

        Parameters
        ----------
        value : float
            The exact values to find.
        possible : array
            The possible grid point values to compare to.

        Returns
        -------
         :
        """
        possible = self.metadata[key]

        # return one index where it's exact
        if value in possible:
            return [value]

        # figure out above and below grid points otherwise
        try:
            above = np.min(possible[possible > value])
        except ValueError:
            above = None
        try:
            below = np.max(possible[possible <= value])
        except ValueError:
            below = None

        if (above is None) or (below is None):
            raise ValueError(
                f"""
            Your requested coordinate of
            {repr(inputs)}
            is outside the limits {np.min(possible)}<={key}<{np.max(possible)}.
            Please rewrite your code to avoid this happening
            or proceed very, very, very, very cautiously.
            """
            )
        else:
            return below, above

    def _get_interpolation_weights(self, value, bounds):
        """
        Get the interpolation weights for `value`
        relative to the `bounds` below and above.
        """

        # if there's only one value, the weight should be 1
        if np.size(bounds) == 1:
            return [1]
        elif np.size(bounds) == 2:
            span = bounds[1] - bounds[0]
            weight_below = (bounds[1] - value) / span
            weight_above = (value - bounds[0]) / span
            return weight_below, weight_above

    _available_metallicities = [-2.0, -1.5, -1.0, -0.5, 0.0, 0.5, 1.0]

    def _wavelengths_to_hashable(self, w):
        """
        Convert an arbitrary array of wavelengths into
        something that can be used as a hash in a dictionary.

        Parameters
        ----------
        w : u.Quantity
            The wavelength array.

        Returns
        -------
        h : tuple
            A static tuple that summarizes the wavelength array
            with its minimum, maximum, and number of elements.
            (Technically this won't be perfectly unique for
            non-uniform grids, but it's probably good enough
            for most uses.)
        """
        return (np.min(w), np.max(w), len(w))

    def _wavelengths_to_R(self, w):
        """
        Calculate the minimum spectral resolution R that
        would be sufficient to support a set of wavelengths.

        Parameters
        ----------
        w : u.Quantity
            The wavelength array.

        Returns
        -------
        R : float
            The maxmium value of w/dw (= 1/dlnw) in the array.
        """
        return 1 / np.min(np.diff(np.log(w.value)))

        return (np.min(w), np.max(w), len(w))

    def _get_average_spectrum_from_grid(
        self, key, wavelength=None, wavelength_edges=None
    ):
        if (wavelength is None) and (wavelength_edges is None):
            return self.models[key]
        else:
            # make sure ask only for one type of wavelength
            assert (wavelength is None) or (wavelength_edges is None)

            # figure out the key associated with these wavelengths
            if wavelength is None:
                wavelength_key = self._wavelengths_to_hashable(wavelength_edges)
            elif wavelength_edges is None:
                wavelength_key = self._wavelengths_to_hashable(wavelength)

            # make sure the dictionary exists
            try:
                self.wavelength_cached_models[wavelength_key]
            except KeyError:
                self.wavelength_cached_models[wavelength_key] = {}

            # get or populate the necessary entry in the grid
            try:
                return self.wavelength_cached_models[wavelength_key][key]
            except KeyError:
                self.wavelength_cached_models[wavelength_key][key] = bintogrid(
                    self.wavelength,
                    self.models[key],
                    newx=wavelength,
                    newx_edges=wavelength_edges,
                )["y"]
                return self.wavelength_cached_models[wavelength_key][key]

    def get_average_spectrum(
        self,
        temperature=5780,
        logg=4.43,
        metallicity=0.0,
        R=100,
        wavelength=None,
        wavelength_edges=None,
        visualize=False,
    ):
        """
        Get a spectrum for an arbitrary temperature, logg, metallicity.

        Parameters
        ----------
        temperature : float
            Temperature, in K (with no astropy units attached).
        logg : float
            Surface gravity log10[g/(cm/s**2)] (with no astropy units attached).
        metallicity : float
            Metallicity log10[metals/solar] (with no astropy units attached).
        R : float
            Spectroscopic resolution (lambda/dlambda). Currently, this must
            be in one of [3,10,30,100,300,1000,3000,10000,30000,100000], but
            check back soon for custom wavelength grids. There is extra
            overhead associated with switching resolutions, so if you're
            going to retrieve many spectra, try to group by resolution.
            (If you're using the `wavelength` or `wavelength_edges` option
            below, please be ensure your requested R exceeds that needed
            to support your wavelengths.)
        wavelength : u.Quantity
            A grid of wavelengths on which you would like your spectrum.
            If this is None, the complete wavelength array will be returned
            at your desired resolution. Otherwise, the spectrum will be
            returned exactly at those wavelengths. Grid points will be
            cached for this new wavelength grid to speed up applications
            that need to retreive lots of similar spectra for the same
            wavelength (like many optimization or sampling problems).
        wavelength_edges : u.Quantity
            Same as `wavelength` (see above!) but defining the wavelength
            grid by its edges instead of its centers. The returned spectrum
            will have 1 fewer element than `wavelength_edges`.

        Returns
        -------
        wavelength : u.Quantity
            The wavelengths, at the specified resolution.
        spectrum : u.Quantity
            The spectrum, in either photons or flux.
        """

        # this kludgy business is to try to avoid loading multiple metallicities unless absolutely necessary
        try:
            # if no wavelength grids are provided, just go with R
            if (wavelength is None) and (wavelength_edges is None):
                assert self.metadata["R"] == R
            # if wavelength grids are provided, make sure the grid has high enough R
            else:
                # make sure ask only for one type of wavelength
                assert (wavelength is None) or (wavelength_edges is None)

                # figure out the minimum R that'd be enough
                if wavelength is None:
                    necessary_R = self._wavelengths_to_R(wavelength_edges)
                elif wavelength_edges is None:
                    necessary_R = self._wavelengths_to_R(wavelength)

                if self.metadata.get("R", 0) < necessary_R:
                    R = self._find_smallest_R(necessary_R)
                    assert False

            assert (metallicity >= np.min(self.metadata["metallicity"])) and (
                metallicity <= np.max(self.metadata["metallicity"])
            )
        except (AttributeError, AssertionError):
            if metallicity in self._available_metallicities:
                self._load_grid(
                    self._find_smallest_R(R=R),
                    metallicity=metallicity,
                )
            else:
                for m in self._available_metallicities:
                    self._load_grid(
                        self._find_smallest_R(R=R),
                        metallicity=m,
                    )

        # store the inputs as a convenient dictionary to pass around
        inputs = dict(temperature=temperature, logg=logg, metallicity=metallicity)

        # figure out the
        bounding_temperature = self._find_bounds(
            temperature, key="temperature", inputs=inputs
        )
        bounding_logg = self._find_bounds(logg, key="logg", inputs=inputs)
        bounding_metallicity = self._find_bounds(
            metallicity, key="metallicity", inputs=inputs
        )

        if visualize:
            self.plot_available(**inputs)

        N = (
            np.size(bounding_temperature)
            * np.size(bounding_logg)
            * np.size(bounding_metallicity)
        )
        spectra = []
        if N == 1:
            weights = [1]
            key = (bounding_temperature[0], bounding_logg[0], bounding_metallicity[0])
            spectrum = self._get_average_spectrum_from_grid(
                key, wavelength=wavelength, wavelength_edges=wavelength_edges
            ).flatten()
        else:
            logT, bounding_logT = np.log(temperature), np.log(bounding_temperature)
            weight_temperature = self._get_interpolation_weights(logT, bounding_logT)
            weight_logg = self._get_interpolation_weights(logg, bounding_logg)
            weight_metallicity = self._get_interpolation_weights(
                metallicity, bounding_metallicity
            )

            weights = np.zeros(N)
            i = 0
            for wt, t in zip(weight_temperature, bounding_temperature):
                for wg, g in zip(weight_logg, bounding_logg):
                    for wz, z in zip(weight_metallicity, bounding_metallicity):
                        weights[i] = wt * wg * wz
                        key = (t, g, z)
                        try:
                            this_log_spectrum = np.log(
                                self._get_average_spectrum_from_grid(
                                    key,
                                    wavelength=wavelength,
                                    wavelength_edges=wavelength_edges,
                                )
                            )
                        except KeyError:
                            raise ValueError(
                                f"""
                            The grid point coordinate {key} is needed to interpolate to your
                            requested coordinate {inputs},
                            but it's not available. This problem is probably caused by
                            requesting something near the jagged edge of the
                            grid's availability.

                            Please run `.plot_available()` to see what models are possible.
                            """
                            )
                        spectra.append(this_log_spectrum)
                        # print(f"{weights[i]} * [{key}]")
                        i += 1
            weight_sum = np.sum(weights)
            assert np.isclose(weight_sum, 1)
            spectrum = np.exp(np.sum(weights[:, np.newaxis] * spectra, axis=0))

        if wavelength is None:
            if wavelength_edges is None:
                wavelength = self.wavelength
            else:
                wavelength = 0.5 * (wavelength_edges[:-1] + wavelength_edges[1:])

        if visualize:
            fi = plt.figure(figsize=(8, 3))
            for w, s in zip(weights, spectra):
                plt.plot(wavelength, s, alpha=w)
            plt.plot(wavelength, spectrum, color="black")
            plt.xlabel(
                f"Wavelength ({self.units['wavelength'].to_string('latex_inline')})"
            )
            plt.ylabel(
                f"Model Flux\n({self.units['spectrum'].to_string('latex_inline')})"
            )
            plt.suptitle(", ".join([f"{k}={v}" for k, v in inputs.items()]))

        return wavelength, spectrum * self.units["spectrum"]

    def plot_available(self, temperature=None, logg=None, metallicity=None):

        N = len(self._keys_for_indexing)
        fi, ax = plt.subplots(
            N, N, figsize=(8, 8), sharex="col", sharey="row", constrained_layout=True
        )
        labels = dict(
            temperature="Temperature (K)",
            logg="$log_{10}[g/(cm/s^2)]$",
            metallicity="[Fe/H] (metallicity)",
        )
        gridkw = dict(marker=".", alpha=0.3)
        for i, ky in enumerate(self._keys_for_indexing):
            for j, kx in enumerate(self._keys_for_indexing):
                y = [k[i] for k in self.models]
                x = [k[j] for k in self.models]
                plt.sca(ax[i, j])
                plt.scatter(x, y, **gridkw)
                plt.scatter(locals()[kx], locals()[ky])
                # if j == 0:
                plt.ylabel(labels[ky])
                # if i == (N - 1):
                plt.xlabel(labels[kx])
        plt.suptitle(f"R={self.metadata['R']}")

    def plot_time_required(self, iterations=5):
        timings = {}
        for k in [
            "R",
            "load library",
            "get spectrum at grid point",
            "get interpolated spectrum",
        ]:
            timings[k] = []

        for R in tqdm(self._available_resolutions, leave=False):
            for i in range(iterations):
                timings["R"].append(R)

                # how long does it take to load?
                start = get_current_seconds()
                self._load_grid(R)
                dt = get_current_seconds() - start
                timings["load library"].append(dt)

                # how long does it take to retrieve a spectrum that exists?
                start = get_current_seconds()
                self.get_average_spectrum(temperature=3000, logg=4.5, metallicity=0.0)
                dt = get_current_seconds() - start
                timings["get spectrum at grid point"].append(dt)

                # how long does it take to retrieve a spectrum that needs to be interpolated?
                start = get_current_seconds()
                self.get_average_spectrum(temperature=3456, logg=5.67, metallicity=0.0)
                dt = get_current_seconds() - start
                timings["get interpolated spectrum"].append(dt)

        t = Table(timings)
        plt.figure(figsize=(8, 4), dpi=300)
        for k in timings.keys():
            if k != "R":
                plt.loglog(timings["R"], timings[k], marker="o", label=k, alpha=0.3)
        plt.legend(bbox_to_anchor=(1, 1), frameon=False)
        plt.xlabel("R = $\lambda/\Delta\lambda$")
        plt.ylabel("Time Required")
        return t


phoenix_library = PHOENIXLibrary(photons=True)
get_phoenix_photons = phoenix_library.get_average_spectrum



from ...imports import *

__all__ = ["from_eureka_S3_txt"]


def eureadka_txt(filename):
    """
    Read eureka's concatenated results table
    and parse it into a Rainbow-friendly format

    Parameters
    ----------
    filename : str
        The path to the file.
    """

    # load the data
    data = ascii.read(filename)

    # figure out a time array
    for time_key in ["time", "bjdtdb"]:
        try:
            t = np.unique(data[time_key])
            break
        except KeyError:
            pass
    timelike = {}
    timelike["time"] = t * u.day * 1

    # figure out a wavelength array
    w = np.unique(data["wave_1d"])
    wavelike = {}
    wavelike["wavelength"] = w * u.micron * 1

    # populate the fluxlike quantities
    fluxlike = {}
    i_time = np.arange(len(t))
    # loop through wavelengths, populating all times for each
    for i_wavelength in tqdm(range(len(w)), leave=False):

        for k in data.colnames[2:]:
            # if an array for this key doesn't exist, create it
            if k not in fluxlike:
                fluxlike[k] = np.zeros((len(w), len(t)))

            # figure out all indices for this wavelengh
            indices_for_this_wavelength = i_wavelength + i_time * len(w)
            fluxlike[k][i_wavelength, i_time] = data[k][indices_for_this_wavelength]

    fluxlike["flux"] = fluxlike["optspec"] * 1
    fluxlike["uncertainty"] = fluxlike["opterr"] * 1

    return wavelike, timelike, fluxlike


def from_eureka_S3_txt(rainbow, filename, **kwargs):
    """
    Populate a Rainbow from a eureka pipeline S3 output.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filename : str
        The path to the file.
    """

    # load the Eureka event
    wavelike, timelike, fluxlike = eureadka_txt(filename)

    # populate the rainbow
    rainbow._initialize_from_dictionaries(
        wavelike=wavelike, timelike=timelike, fluxlike=fluxlike
    )



from ...imports import *
from ...spectra import *

__all__ = ["inject_spectrum"]


def inject_spectrum(
    self,
    temperature=5800 * u.K,
    logg=4.43,
    metallicity=0.0,
    radius=1 * u.Rsun,
    distance=10 * u.pc,
    phoenix=True,
):
    """
    Inject a stellar spectrum.

    Parameters
    ----------
    temperature : u.Quantity
        Temperature, in K (with no astropy units attached).
    logg : float
        Surface gravity log10[g/(cm/s**2)] (with no astropy units attached).
    metallicity : float
        Metallicity log10[metals/solar] (with no astropy units attached).
    radius : u.Quantity
        The radius of the star.
    distance : u.Quantity
        The distance to the star.
    phoenix : bool
        If `True`, use PHOENIX surface flux.
        If `False`, use Planck surface flux.

    Returns
    -------
    rainbow : Rainbow
        A new Rainbow object with the spectrum injected.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inject_spectrum", locals())

    # create a copy of the existing Rainbow
    new = self._create_copy()

    # warn if maybe we shouldn't inject anything
    if np.all(self.flux != 1):
        warnings.warn(
            f"""
        None of the pre-existing flux values were 1,
        which hints at the possibility that there
        might already be a spectrum in them. Please
        watch out for weird units or values!
        """
        )

    if phoenix:
        f = get_phoenix_photons
    else:
        f = get_planck_photons

    # get the spectrum from the surface
    _, surface_flux = f(
        temperature=u.Quantity(temperature).value,
        logg=logg,
        metallicity=metallicity,
        wavelength=self.wavelength,
    )

    # get the received flux at Earth
    received_flux = surface_flux * (radius / distance).decompose() ** 2

    # do math with spectrum
    for k in ["flux", "model", "uncertainty"]:
        try:
            new.fluxlike[k] = self.get(k) * self._broadcast_to_fluxlike(received_flux)
        except KeyError:
            pass

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new object
    return new



from ....imports import *

__all__ = ["get_average_lightcurve"]


def get_average_lightcurve(self):
    """
    Return a lightcurve of the star, averaged over all wavelengths.

    Parameters
    ----------

    Returns
    -------
    lightcurve : np.array (wavelike)
    """
    return self.get_average_lightcurve_as_rainbow().flux[0, :]



"""
Define a reader for STScI pipeline x1dints.fits files.
"""
from ...imports import *

__all__ = ["from_x1dints_kludge"]


def setup_integration_times(filenames):
    """
    Figure out the times of each integration,
    either by reading them from the headers
    or by totally just making them up from scratch.

    Parameters
    ----------
    filenames : list
        A sorted list of filenames, corresponding to one or more segments.
    """

    # create an empty timelike dictionary
    timelike = {}

    # check the first segment
    first_hdu = fits.open(filenames[0])

    # does the first segment define some times?
    try:
        assert len(first_hdu["int_times"].data) > 0

        # grab the integration time information
        for c in first_hdu["int_times"].data.columns.names:
            timelike[c] = first_hdu["int_times"].data[c]

        # be sure to set our standard time axis
        timelike["time"] = timelike["int_mid_BJD_TDB"] * u.day
    # if times are not in header, make up some imaginary ones!
    except:
        # alert the user to what we're doing
        warnings.warn("No times found! Making up imaginary ones!")
        last_hdu = fits.open(filenames[-1])

        # figure out the total number of integrations (DOES THIS NEED THE -1?)
        try:
            N_integrations = last_hdu["PRIMARY"].header["INTEND"] - 1
        except KeyError:
            # (this kludge necessary for CV-simulated NIRSpec)
            N_integrations = last_hdu["PRIMARY"].header["NINTS"]

        # get the time per integration (DOES THIS INCLUDE OVERHEADS?)
        time_per_integration = last_hdu["PRIMARY"].header["EFFINTTM"] * u.s
        warnings.warn(f"The imaginary times assume {time_per_integration}/integration.")

        # create a fake array of times
        fake_times = np.arange(N_integrations) * time_per_integration
        timelike["time"] = fake_times.to(u.day)

    # return a timelike dictionary
    return timelike


def from_x1dints_kludge(rainbow, filepath, **kw):
    """
    Populate a Rainbow from an STScI pipeline x1dints file,
    or a group of x1dints files for multiple segments.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated. (This is intended
        to be used as a class method of Rainbow or
        of a class derived from Rainbow, as a way of
        initializing an object from files.)
    filepath : str
        The path to the file, or a glob-friendly string
        pointing to a group of files that should all be
        loaded together (for example, if an exposure was
        split into multiple segments).
    """

    # create a list of filenames (which might be just 1)
    filenames = expand_filenames(filepath)

    # figure out the times (necessary to set up arrays)
    timelike = setup_integration_times(filenames)
    rainbow.timelike.update(**timelike)

    # loop over file (each one a segment)
    for i_file, f in enumerate(tqdm(filenames, leave=False)):

        # open this fits file
        hdu = fits.open(f)

        # if this is the first one, populate the shared stuff
        if i_file == 0:

            # grab the entire primary header
            rainbow.metadata["header"] = hdu["PRIMARY"].header
            # TO-DO: (watch out for the things that aren't constant across segments!)

        # set the index to the start of this segment
        try:
            integration_counter = hdu["PRIMARY"].header["INTSTART"]
        except KeyError:
            # (this kludge necessary for CV-simulated NIRSpec)
            integration_counter = 0

        try:
            n_integrations_predicted_by_header = (
                hdu["PRIMARY"].header["INTEND"] - hdu["PRIMARY"].header["INTSTART"]
            )
        except KeyError:
            n_integrations_predicted_by_header = 0
        n_integrations_in_segment = np.maximum(
            len(hdu) - 3, n_integrations_predicted_by_header
        )

        # print(integration_counter, n_integrations_in_segment, len(hdu))

        # loop through the spectra
        for e in range(2, 2 + n_integrations_in_segment):

            # do this stuff only on the very first time through
            if i_file == 0 and e == 2:
                # pull out a wavelength grid (assuming it'll stay constant)
                unit_string = hdu[e].columns["wavelength"].unit
                if unit_string is None:
                    unit_string = ""
                wavelength_unit = u.Unit(unit_string)
                rainbow.wavelike["wavelength"] = (
                    hdu[e].data["wavelength"] * wavelength_unit * 1
                )

                # set up the fluxlike quantities
                column_units = {}
                for column in hdu[e].columns:

                    # get a lower case name for the unit
                    c = column.name.lower()

                    # set up an empty Quantity array, with units if known
                    this_quantity = u.Quantity(
                        np.nan * np.ones((rainbow.nwave, rainbow.ntime))
                    )
                    if column.unit is not None:
                        column_units[c] = u.Unit(column.unit)
                        this_quantity *= column_units[c]
                    else:
                        column_units[c] = 1

                    # populate the fluxlike dictionary with the empty array
                    rainbow.fluxlike[c] = this_quantity

            for column in hdu[e].columns:

                # get a lower case name for the unit
                c = column.name.lower()
                rainbow.fluxlike[c][:, integration_counter - 1] = (
                    hdu[e].data[c] * column_units[c]
                )

            # increment the running integration total
            integration_counter += 1

        # when done with this file, make sure indices line up
        # assert integration_counter == hdu["PRIMARY"].header["INTEND"]
        #  FIXME - it looks like different instruments may have different INTENDs?
    n_filled_times = np.sum(np.any(np.isfinite(rainbow.flux), rainbow.waveaxis))
    if n_filled_times != rainbow.ntime:
        warnings.warn(
            f"""
        The x1dints header(s) indicate there should be {rainbow.ntime} integrations,
        but only {n_filled_times} columns of the flux array were populated. Are you
        perhaps missing some segment files?
        """
        )

    # try to pull in the errors
    try:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["flux_error"] * 1
    except KeyError:
        rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["error"] * 1

    if rainbow.uncertainty is None:
        message = f"""
        Hmmm...it's not clear which column corresponds to the
        flux uncertainties for this Rainbow object. The
        available `fluxlike` columns are:
            {rainbow.fluxlike.keys()}
        A long-term solution might be to fix the `from_x1dints_kludge`
        reader, but a short-term solution would be to pick one
        of the columns listed above and say something like

        x.fluxlike['uncertainty'] = x.fluxlike['some-other-relevant-error-column']

        where `x` is the Rainbow you just created.
        """
        warnings.warn(message)

    # try to guess wscale (and then kludge and call it linear)
    # rainbow._guess_wscale()
    # rainbow.metadata['wscale'] = 'linear' # TODO: fix this kludge
